"""
This module contains the KMLWriter class.
"""

import os
import warnings
import xml.etree.ElementTree as ET
from typing import Optional

from ..gpx_elements import Gpx
from .writer import Writer

DEFAULT_NORMAL_STYLE = {"color": "ff0000ff", "width": 2, "fill": 0}

DEFAULT_HIGHLIGHT_STYLE = {"color": "ff0000ff", "width": 2, "fill": 0}

DEFAULT_STYLES = [
    ("normal", DEFAULT_NORMAL_STYLE),
    ("highlight", DEFAULT_HIGHLIGHT_STYLE),
]


class KMLWriter(Writer):
    """
    KML file writer.
    """

    def __init__(
        self,
        gpx: Gpx = None,
        *,
        properties: bool = True,
        metadata: bool = True,  # unused
        waypoints: bool = True,
        routes: bool = True,
        extensions: bool = True,
        ele: bool = True,
        time: bool = True,
        precisions: dict = None,
        time_format: str = None,
        styles: list[tuple[str, dict]] = None,
    ) -> None:
        """
        Initialise GPXWriter instance.
        """
        super().__init__(gpx, precisions, time_format)

        # Parameters
        self.properties: bool = properties
        self.metadata: bool = metadata
        self.waypoints: bool = waypoints
        self.routes: bool = routes
        self.extensions: bool = extensions
        self.ele: bool = ele
        self.time: bool = time

        self.styles = DEFAULT_STYLES if styles is None else styles

        # Utility attributes
        self.kml_string: str = ""
        self.kml_root = None

    def add_pair(self, element: ET.Element, key: str, style_url: str) -> ET.Element:
        """
        Add StyleMap to KML element.

        Args:
            element (ET.Element): KML element.
            key (str): Key.
            style_url (str): Style URL.

        Returns:
            ET.Element: KML element.
        """
        pair_ = ET.SubElement(element, "Pair")
        pair_, _ = self.add_subelement(pair_, "key", key)
        pair_, _ = self.add_subelement(pair_, "styleUrl", style_url)
        return element

    def add_stylemap(self, element: ET.Element, id_: str) -> ET.Element:
        """
        Add StyleMap to KML element.

        Args:
            element (ET.Element): KML element.
            id_ (str): StyleMap element id.

        Returns:
            ET.Element: KML element.
        """
        stylemap_ = ET.SubElement(element, "StyleMap")
        self.set_not_none(stylemap_, "id", id_)
        style_id = 1
        for style_key, _ in self.styles:  # _ used to be called style
            stylemap_ = self.add_pair(stylemap_, style_key, "#style" + str(style_id))
            style_id += 1
        return element

    def add_polystyle(self, element: ET.Element, style: dict) -> ET.Element:
        """
        Add PolyStyle to KML element.

        Args:
            element (ET.Element): KML element.
            style (dict): Style.

        Returns:
            ET.Element: KML element.
        """
        polystyle_ = ET.SubElement(element, "PolyStyle")
        try:
            polystyle_, _ = self.add_subelement_number(
                polystyle_, "fill", style["fill"]
            )
        except KeyError:
            warnings.warn("No fill attribute in style")
            polystyle_, _ = self.add_subelement_number(polystyle_, "fill", 0)
        return element

    def add_linestyle(self, element: ET.Element, style: dict) -> ET.Element:
        """
        Add LineStyle to KML element.

        Args:
            element (ET.Element): KML element.
            style (dict): Style.

        Returns:
            ET.Element: KML element.
        """
        linestyle_ = ET.SubElement(element, "LineStyle")
        try:
            linestyle_, _ = self.add_subelement(linestyle_, "color", style["color"])
        except KeyError:
            warnings.warn("No color attribute in style")
            linestyle_, _ = self.add_subelement(linestyle_, "color", "ff0000ff")
        try:
            linestyle_, _ = self.add_subelement_number(
                linestyle_, "width", style["width"]
            )
        except KeyError:
            warnings.warn("No width attribute in style")
            linestyle_, _ = self.add_subelement_number(linestyle_, "width", 2)
        return element

    def add_style(
        self,
        element: ET.Element,
        id: str,  # pylint: disable=redefined-builtin
        style: dict,
    ) -> ET.Element:
        """
        Add Style to KML element.

        Args:
            element (ET.Element): KML element.
            id (str): Style element id.
            style (dict): Line style.

        Returns:
            ET.Element: KML element.
        """
        style_ = ET.SubElement(element, "Style")
        self.set_not_none(style_, "id", id)
        style_ = self.add_linestyle(style_, style)
        style_ = self.add_polystyle(style_, style)
        return element

    def add_linestring(self, element: ET.Element) -> ET.Element:
        """
        Add LineString to KML element.

        Args:
            element (ET.Element): KML element.

        Returns:
            ET.Element: KML element.
        """
        linestring_ = ET.SubElement(element, "LineString")
        linestring_, _ = self.add_subelement_number(linestring_, "tessellate", 1)
        if self.ele:
            coordinates = self.gpx.to_csv(
                dest=None, values=["lon", "lat", "ele"]
            ).replace("\n", " ")
        else:
            coordinates = self.gpx.to_csv(dest=None, values=["lon", "lat"]).replace(
                "\n", " "
            )
        linestring_, _ = self.add_subelement(linestring_, "coordinates", coordinates)
        return element

    def add_placemark(self, element: ET.Element) -> ET.Element:
        """
        Add Placemark to KML element.

        Args:
            element (ET.Element): KML element.

        Returns:
            ET.Element: KML element.
        """
        placemark_ = ET.SubElement(element, "Placemark")
        placemark_, _ = self.add_subelement(placemark_, "name", self.gpx.name())
        placemark_, _ = self.add_subelement(placemark_, "styleUrl", "#stylemap")
        placemark_ = self.add_linestring(placemark_)
        return element

    def add_document(self, element: ET.Element) -> ET.Element:
        """
        Add Document to KML element.

        Args:
            element (ET.Element): KML element.

        Returns:
            ET.Element: KML element.
        """
        document_ = ET.SubElement(element, "Document")
        document_, _ = self.add_subelement(
            document_, "name", os.path.basename(self.file_path)
        )
        id_ = 1
        for _, style in self.styles:  # _ used to be called style_key
            document_ = self.add_style(document_, "style" + str(id), style)
            id_ += 1
        document_ = self.add_stylemap(document_, "stylemap")
        document_ = self.add_placemark(document_)
        return element

    def add_root_document(self) -> None:
        """
        Add Document element to the KML root element.
        """
        self.kml_root = self.add_document(self.kml_root)

    def add_root_properties(self) -> None:
        """
        Add properties to the GPX root element.
        """
        # According to .kml file from Google Earth Pro
        self.kml_root.set("xmlns", "http://www.opengis.net/kml/2.2")
        self.kml_root.set("xmlns:gx", "http://www.google.com/kml/ext/2.2")
        self.kml_root.set("xmlns:kml", "http://www.opengis.net/kml/2.2")
        self.kml_root.set("xmlns:atom", "http://www.w3.org/2005/Atom")

    def gpx_to_string(self) -> str | None:
        """
        Convert Gpx instance to a string (the content of a .kml file).

        Returns:
            str | None: String corresponding to the Gpx instance.
        """
        if self.gpx is None:
            return None

        # Reset string
        self.kml_string = ""

        # Root
        self.kml_root = ET.Element("kml")

        # Properties
        if self.properties:
            self.add_root_properties()

        # Document
        self.add_root_document()

        # Convert data to string
        self.kml_string = ET.tostring(self.kml_root, encoding="unicode")
        # self.kml_string = ET.tostring(kml_root, encoding="utf-8")

        return self.kml_string

    def write_gpx(self):
        """
        Convert Gpx instance to string and write to .kml file.
        """
        # Open/create KML file
        try:
            f = open(self.file_path, "w", encoding="utf-8")
        except OSError:
            warnings.warn("Could not open/read file: %s", self.file_path)
            raise
        # Write KML file
        with f:
            f.write('<?xml version="1.0" encoding="UTF-8"?>')
            f.write(self.kml_string)

    def write(
        self,
        file_path: str,
        styles: Optional[list[tuple[str, dict]]] = None,
        xml_schema: bool = False,
        xml_extensions_schemas: bool = False,
    ) -> bool:
        """
        Handle writing.

        Args:
            file_path (str): Path to write the KML file.
            styles (Optional[list[tuple[str, dict]]], optional): List of
                (style_id, style) tuples. Defaults to None.
            xml_schema (bool, optional): Toggle schema verification
                after writting. Defaults to False.
            xml_extensions_schemas (bool, optional): TODO.
                Defaults to False.

        Returns:
            bool: Return True if written file follows checked schemas,
                False otherwise.
        """
        # Handle path
        directory_path = os.path.dirname(os.path.realpath(file_path))
        if not os.path.exists(directory_path):
            warnings.warn("Provided path does not exist")
            return False
        self.file_path = file_path

        # Update style
        if styles is not None:
            self.styles = styles

        # Write .kml file
        self.gpx_to_string()
        self.write_gpx()

        # Check XML schemas
        res = True
        if xml_schema or xml_extensions_schemas:
            res = self.xml_schemas(xml_schema, xml_extensions_schemas)

        return res



"""
This module contains the Email class.
"""

from .gpx_element import GpxElement


class Email(GpxElement):
    """
    emailType element in GPX file.
    """

    fields = ["id", "domain"]
    mandatory_fields = ["id", "domain"]

    def __init__(
        self,
        tag: str = "email",
        id: str = None,  # pylint: disable=redefined-builtin
        domain: str = None,
    ) -> None:
        """
        Initialise Email instance.

        Args:
            tag (str, optional): XML tag. Defaults to "email".
            id (str, optional): Identifier. Defaults to None.
            domain (str, optional): Domain. Defaults to None.
        """
        self.tag: str = tag
        self.id: str = id
        self.domain: str = domain



"""
This module contains the Track class.
"""

from .extensions import Extensions
from .gpx_element import GpxElement
from .link import Link
from .track_segment import TrackSegment


class Track(GpxElement):
    """
    trkType element in GPX file.
    """

    fields = ["name", "cmt", "desc", "link", "number", "type", "extensions", "trkseg"]
    mandatory_fields = []

    def __init__(
        self,
        tag: str = "trk",
        name: str = None,
        cmt: str = None,
        desc: str = None,
        src: str = None,
        link: Link = None,
        number: int = None,
        type: str = None,  # pylint: disable=redefined-builtin
        extensions: Extensions = None,
        trkseg: list[TrackSegment] = None,
    ) -> None:
        """
        Initialise Track instance.

        Args:
            tag (str, optional): XML tag. Defaults to "trk".
            name (str, optional): Name. Defaults to None.
            cmt (str, optional): Comment. Defaults to None.
            desc (str, optional): Description. Defaults to None.
            src (str, optional): Source. Defaults to None.
            link (Link, optional): Link. Defaults to None.
            number (int, optional): Number. Defaults to None.
            type (str, optional): Type. Defaults to None.
            extensions (Extensions, optional): Extensions. Defaults to None.
            trkseg (list[TrackSegment], optional): List of track
                segments. Defaults to None.
        """
        self.tag: str = tag
        self.name: str = name
        self.cmt: str = cmt
        self.desc: str = desc
        self.src: str = src
        self.link: Link = link
        self.number: int = number
        self.type: str = type
        self.extensions: Extensions = extensions
        self.trkseg: list[TrackSegment] = [] if trkseg is None else trkseg



"""
This module contains the Link class.
"""

from .gpx_element import GpxElement


class Link(GpxElement):
    """
    linkType element in GPX file.
    """

    fields = ["href", "text", "type"]
    mandatory_fields = ["href"]

    def __init__(
        self,
        tag: str = "link",
        href: str = None,
        text: str = None,
        type: str = None,  # pylint: disable=redefined-builtin
    ) -> None:
        """
        Initialise Link instance.

        Args:
            tag (str, optional): XML tag. Defaults to "link".
            href (str, optional): Hyper reference. Defaults to None.
            text (str, optional): Text. Defaults to None.
            type (str, optional): Type. Defaults to None.
        """
        self.tag: str = tag
        self.href: str = href
        self.text: str = text
        self.type: str = type



"""
This module contains the FitParser class.
"""

from pathlib import Path
from typing import IO

from fitparse import FitFile

from ..constants.precisions import DEFAULT_PRECISION
from ..gpx_elements import Gpx, Track, TrackSegment, WayPoint
from .parser import Parser


class FitParser(Parser):
    """
    Fit file parser.
    """

    _semicircles_to_deg_const = 180 / 2**31

    def __init__(self, source: str | Path | IO[str] | IO[bytes] | bytes) -> None:
        """
        Initialise FitParser instance.

        Args:
            source (str | Path | IO[str] | IO[bytes] | bytes): Path to a
                file or a file-like object to parse.
        """
        # Initialise Parser and parse FIT file
        super().__init__(source)
        self.parse()

    def _semicircles_to_deg(self, semicircle_coord: list) -> list:
        """
        Convert semicircle data from FIT file to dms data.

        Args:
            semicircle_coord (list): list of semicircle values.

        Returns:
            list: list of dms values.
        """
        return [FitParser._semicircles_to_deg_const * x for x in semicircle_coord]

    def _parse(self):
        """
        Parse FIT file and store data in a Gpx element.
        """
        lat_data = []
        lon_data = []
        alt_data = []
        time_data = []

        units = {"alt": "", "lat": "", "lon": ""}

        fit_file = FitFile(self.source)

        for record in fit_file.get_messages("record"):
            for record_data in record:
                if record_data.name == "position_lat":
                    lat_data.append(record_data.value)
                    if units["lat"] == "":
                        units["lat"] = record_data.units
                        # Temporary (may change if units["lat"] == "semicircles")
                        self.precisions["lat_lon"] = self._find_precision(
                            record_data.value
                        )
                if record_data.name == "position_long":
                    lon_data.append(record_data.value)
                    if units["lon"] == "":
                        units["lon"] = record_data.units
                if record_data.name == "altitude":
                    alt_data.append(record_data.value)
                    if units["alt"] == "":
                        units["alt"] = record_data.units
                        self.precisions["elevation"] = self._find_precision(
                            record_data.value
                        )
                if record_data.name == "timestamp":
                    time_data.append(record_data.value)
                    self._find_time_format(record_data.value)

        # Convert semicircles data to radians ??
        if units["lat"] == "semicircles":
            self.precisions["lat_lon"] = DEFAULT_PRECISION
            lat_data = self._semicircles_to_deg(lat_data)
        if units["lon"] == "semicircles":
            lon_data = self._semicircles_to_deg(lon_data)

        # Store FIT data in Gpx element
        trkpt = [
            WayPoint("trkpt", lat, lon, alt, time)
            for lat, lon, alt, time in zip(lat_data, lon_data, alt_data, time_data)
        ]
        trkseg = TrackSegment(trkpt=trkpt)
        trk = Track(trkseg=[trkseg])
        self.gpx.trk = [trk]

    def _add_properties(self):
        self.gpx.creator = "ezGPX"
        self.gpx.xmlns = "http://www.topografix.com/GPX/1/1"
        self.gpx.version = "1.1"
        self.gpx.xmlns_xsi = "http://www.w3.org/2001/XMLSchema-instance"
        self.gpx.xsi_schema_location = [
            "http://www.topografix.com/GPX/1/1",
            "http://www.topografix.com/GPX/1/1/gpx.xsd",
        ]

    def parse(self) -> Gpx:
        """
        Parse Fit file.

        Returns:
            Gpx: Gpx instance.
        """
        # Parse FIT file
        self._parse()

        # Add properties
        self._add_properties()

        return self.gpx



"""
This module contains algorithms.
"""

from math import degrees

from .distance import EARTH_RADIUS, perpendicular_distance


def ramer_douglas_peucker(
    points: list, epsilon: float = degrees(2 / EARTH_RADIUS)
) -> list:
    """
    Simplify a curve using the Ramer-Douglas-Peucker algorithm.
    Source: https://en.wikipedia.org/wiki/Ramer%E2%80%93Douglas%E2%80%93Peucker_algorithm

    Args:
        points (list): List of points defining the track to simplify.
        epsilon (float, optional): Ramer-Douglas-Peucker threshold
            distance (higher value means more simplifications).
            Defaults to degrees(2 / EARTH_RADIUS), (i.e.: the angle
            corresponding to a distance of 2 metres at the surface of
            the earth).

    Returns:
        list: List of points defining the simplified track.
    """
    # Find the point with the maximum distance
    d_max = 0
    i_max = 0

    start_point = points[0]
    end_point = points[len(points) - 1]

    for i in range(1, len(points) - 1):
        d = perpendicular_distance(start_point, end_point, points[i])
        if d > d_max:
            d_max = d
            i_max = i

    result = []

    # If max distance is greater than epsilon, recursively simplify
    if d_max > epsilon:
        # Recursive call
        result_1 = ramer_douglas_peucker(points[0 : i_max + 1], epsilon)
        result_2 = ramer_douglas_peucker(points[i_max : len(points)], epsilon)

        # Build result list
        result = result_1 + result_2[1:]
    else:
        result = [start_point, end_point]

    return result



"""
This module contains constants related to precision.
"""

DEFAULT_PRECISION = 10
DEFAULT_PRECISION_DICT = {
    "lat_lon": DEFAULT_PRECISION,
    "elevation": DEFAULT_PRECISION,
    "distance": DEFAULT_PRECISION,
    "duration": DEFAULT_PRECISION,
    "speed": DEFAULT_PRECISION,
    "rate": DEFAULT_PRECISION,
    "default": DEFAULT_PRECISION,
}
POSSIBLE_TIME_FORMATS = ["%Y-%m-%dT%H:%M:%SZ", "%Y-%m-%dT%H:%M:%S.%fZ"]
DEFAULT_TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"



"""
This module contains the Copyright class.
"""

from .gpx_element import GpxElement


class Copyright(GpxElement):
    """
    copyrightType element in GPX file.
    """

    fields = ["author", "year", "licence"]
    mandatory_fields = ["author"]

    def __init__(
        self,
        tag: str = "copyright",
        author: str = None,
        year: int = None,
        licence: str = None,
    ) -> None:
        """
        Initialise Copyright instance.

        Args:
            tag (str, optional): XML tag. Defaults to "copyright".
            author (str, optional): Author. Defaults to None.
            year (int, optional): Year. Defaults to None.
            licence (str, optional): Licence. Defaults to None.
        """
        self.tag: str = tag
        self.author: str = author
        self.year: int = year
        self.licence: str = licence



"""
This module contains the Gpx class.
"""

from __future__ import annotations

import io
import warnings
from datetime import datetime, timezone
from pathlib import Path
from typing import IO, Any, Optional, Tuple

import narwhals as nw
import pandas as pd
import polars as pl
from narwhals.typing import IntoFrameT

from ..utils import haversine_distance, ramer_douglas_peucker
from .extensions import Extensions
from .gpx_element import GpxElement
from .metadata import Metadata
from .route import Route
from .track import Track
from .track_segment import TrackSegment
from .waypoint import WayPoint


class Gpx(GpxElement):
    """
    gpxType element in GPX file.
    """

    fields = ["version", "creator", "metadata", "wpt", "rte", "trk", "extensions"]
    mandatory_fields = ["version", "creator"]

    def __init__(
        self,
        *,
        tag: str = "gpx",
        version: str = None,
        creator: str = None,
        xsi_schema_location: list[str] = None,
        xmlns: dict = None,
        metadata: Metadata = None,
        wpt: list[WayPoint] = None,
        rte: list[Route] = None,
        trk: list[Track] = None,
        extensions: Extensions = None,
    ) -> None:
        """
        Initialise Gpx instance.

        Args:
            tag (str, optional): XML tag. Defaults to "gpx".
            version (str, optional): GPX schema version. Defaults to None.
            creator (str, optional): Creator. Defaults to None.
            xsi_schema_location (list[str], optional): XML schema
                instance. Defaults to None.
            xmlns (dict, optional): XML name spaces. Defaults to None.
            metadata (Metadata, optional): Metadata. Defaults to None.
            wpt (list[WayPoint], optional): List of waypoints. Defaults to None.
            rte (list[Route], optional): List of routes. Defaults to None.
            trk (list[Track], optional): List of tracks. Defaults to None.
            extensions (Extensions, optional): Extensions. Defaults to None.
        """
        self.tag: str = tag
        self.version: str = version
        self.creator: str = creator
        self.xsi_schema_location: list[str] = (
            [] if xsi_schema_location is None else xsi_schema_location
        )
        self.xmlns: dict = {} if xmlns is None else xmlns
        self.metadata: Metadata = metadata
        self.wpt: list[WayPoint] = [] if wpt is None else wpt
        self.rte: list[Route] = [] if rte is None else rte
        self.trk: list[Track] = [] if trk is None else trk
        self.extensions: Extensions = extensions

        # Markers
        self._dist_from_start = False
        self.ascent_rate = False
        self._speed: bool = False
        self._pace: bool = False
        self._ascent_speed: bool = False

    @staticmethod
    def from_dataframe(df: IntoFrameT) -> Gpx:
        """
        Initialize Gpx from dataframe.

        Args:
            df (IntoFrameT): Dataframe with "lat", "lon" columns. Also
                supports "ele", "time" columns.
        """
        df = nw.from_native(df)

        trkpt = [
            WayPoint(
                tag="trkpt",
                lat=row["lat"],
                lon=row["lon"],
                ele=row.get("ele"),
                time=row.get("time"),
            )
            for row in df.iter_rows(named=True)
        ]
        trkseg = TrackSegment(trkpt=trkpt)
        trk = Track(trkseg=[trkseg])
        return Gpx(trk=[trk])

    ###############################################################################
    #### Name #####################################################################
    ###############################################################################

    def name(self) -> str:
        """
        Return activity name.
        """
        return self.trk[0].name

    def set_name(self, new_name: str) -> None:
        """
        Set name.

        Args:
            new_name (str): New name.
        """
        self.trk[0].name = new_name

    ###############################################################################
    #### Points ###################################################################
    ###############################################################################

    def nb_points(self) -> int:
        """
        Return the number of points in the GPX.
        """
        nb_pts = 0
        for track in self.trk:
            for track_segment in track.trkseg:
                nb_pts += len(track_segment.trkpt)
        return nb_pts

    def bounds(self) -> Tuple[float, float, float, float]:
        """
        Find minimum and maximum latitude and longitude.

        Returns:
            Tuple[float, float, float, float]: Min latitude, min
                longitude, max latitude, max longitude.
        """
        min_lat = self.trk[0].trkseg[0].trkpt[0].lat
        min_lon = self.trk[0].trkseg[0].trkpt[0].lon
        max_lat = min_lat
        max_lon = min_lon

        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    min_lat = min(min_lat, track_point.lat)
                    min_lon = min(min_lon, track_point.lon)
                    max_lat = max(max_lat, track_point.lat)
                    max_lon = max(max_lon, track_point.lon)
        return min_lat, min_lon, max_lat, max_lon

    def center(self) -> Tuple[float, float]:
        """
        Return latitude and longitude of the center point.
        """
        min_lat, min_lon, max_lat, max_lon = self.bounds()
        center_lat = min_lat + (max_lat - min_lat) / 2
        center_lon = min_lon + (max_lon - min_lon) / 2
        return center_lat, center_lon

    def first_point(self) -> WayPoint:
        """
        Return GPX first point.
        """
        return self.trk[0].trkseg[0].trkpt[0]

    def last_point(self) -> WayPoint:
        """
        Return GPX last point.
        """
        return self.trk[-1].trkseg[-1].trkpt[-1]

    def extreme_points(self) -> Tuple[WayPoint, WayPoint, WayPoint, WayPoint]:
        """
        Find extreme points in track, i.e.: points with lowest and
        highest latitude and longitude.

        Returns:
            Tuple[WayPoint, WayPoint, WayPoint, WayPoint]: Min latitude
                point, min longitude point, max latitude point, max
                longitude point.
        """
        min_lat_point = self.trk[0].trkseg[0].trkpt[0]
        min_lon_point = self.trk[0].trkseg[0].trkpt[0]
        max_lat_point = self.trk[0].trkseg[0].trkpt[0]
        max_lon_point = self.trk[0].trkseg[0].trkpt[0]

        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.lat < min_lat_point.lat:
                        min_lat_point = track_point
                    if track_point.lon < min_lon_point.lon:
                        min_lon_point = track_point
                    if track_point.lat > max_lat_point.lat:
                        max_lat_point = track_point
                    if track_point.lon > max_lon_point.lon:
                        max_lon_point = track_point
        return min_lat_point, min_lon_point, max_lat_point, max_lon_point

    ###############################################################################
    #### Distance and Elevation ###################################################
    ###############################################################################

    def distance(self) -> float:
        """
        Return the total distance (in meters) of tracks contained in
        the Gpx element.
        """
        dst = 0.0
        previous_point = self.trk[0].trkseg[0].trkpt[0]
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    dst += haversine_distance(previous_point, track_point)
                    previous_point = track_point
        return dst

    def _compute_distance_from_start(self):
        """
        Return distance from start at each point.
        """
        if self._dist_from_start:
            return
        dst = 0.0
        previous_point = self.trk[0].trkseg[0].trkpt[0]
        previous_point.distance_from_start = dst
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    dst += haversine_distance(previous_point, track_point)
                    track_point.distance_from_start = dst
                    previous_point = track_point
        self._dist_from_start = True

    def ascent(self) -> float:
        """
        Return the total ascent (in meters) of tracks contained in the
        Gpx element.
        """
        ascent = 0
        previous_elevation = self.trk[0].trkseg[0].trkpt[0].ele
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.ele > previous_elevation:
                        ascent += track_point.ele - previous_elevation
                    previous_elevation = track_point.ele
        return ascent

    def descent(self) -> float:
        """
        Return the total descent (in meters) of tracks contained in the
        Gpx element.
        """
        descent = 0
        previous_elevation = self.trk[0].trkseg[0].trkpt[0].ele
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.ele < previous_elevation:
                        descent += previous_elevation - track_point.ele
                    previous_elevation = track_point.ele
        return descent

    def min_elevation(self) -> float:
        """
        Return minimum elevation (in meters) in tracks contained in the
        Gpx element.
        """
        min_elevation = self.trk[0].trkseg[0].trkpt[0].ele
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    min_elevation = min(min_elevation, track_point.ele)
        return min_elevation

    def max_elevation(self) -> float:
        """
        Return maximum elevation (in meters) in tracks contained in the
        Gpx element.
        """
        max_elevation = self.trk[0].trkseg[0].trkpt[0].ele
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    max_elevation = max(max_elevation, track_point.ele)
        return max_elevation

    def _compute_ascent_rate(self) -> None:
        """
        Compute ascent rate at each point.
        """
        if self.ascent_rate:
            return
        previous_point = self.first_point()
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    distance = haversine_distance(previous_point, track_point)
                    ascent = track_point.ele - previous_point.ele
                    try:
                        track_point.ascent_rate = (ascent * 100) / distance
                    except ZeroDivisionError:
                        track_point.ascent_rate = 0.0
                    previous_point = track_point
        self.ascent_rate = True

    def max_descent_rate(self) -> float:
        """
        Return activity minimum ascent rate.
        """
        self._compute_ascent_rate()
        min_ascent_rate = 100.0
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    min_ascent_rate = min(min_ascent_rate, track_point.ascent_rate)
        return min_ascent_rate

    def max_ascent_rate(self) -> float:
        """
        Return activity maximum ascent rate.
        """
        self._compute_ascent_rate()
        max_ascent_rate = -1.0
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    max_ascent_rate = max(max_ascent_rate, track_point.ascent_rate)
        return max_ascent_rate

    ###############################################################################
    #### Time #####################################################################
    ###############################################################################

    def utc_start_time(self) -> datetime:
        """
        Return the activity UTC start time.
        """
        return self.trk[0].trkseg[0].trkpt[0].time

    def utc_stop_time(self):
        """
        Return the activity UTC stop time.
        """
        return self.trk[-1].trkseg[-1].trkpt[-1].time

    def start_time(self) -> datetime:
        """
        Return the activity start time.
        """
        start_time = None
        try:
            start_time = (
                self.trk[0]
                .trkseg[0]
                .trkpt[0]
                .time.replace(tzinfo=timezone.utc)
                .astimezone(tz=None)
            )
        except AttributeError:
            warnings.warn("Unable to find activity start time")
        return start_time

    def stop_time(self) -> datetime:
        """
        Return the activity stop time.
        """
        stop_time = None
        try:
            stop_time = (
                self.trk[-1]
                .trkseg[-1]
                .trkpt[-1]
                .time.replace(tzinfo=timezone.utc)
                .astimezone(tz=None)
            )
        except AttributeError:
            warnings.warn("Unable to find activity stop time")
        return stop_time

    def total_elapsed_time(self) -> datetime:
        """
        Return the total elapsed time.
        """
        total_elapsed_time = None
        try:
            total_elapsed_time = self.stop_time() - self.start_time()
        except TypeError:
            warnings.warn("Unable to compute activity total elapsed time")
        return total_elapsed_time

    def stopped_time(self, tolerance: float = 2.45) -> datetime:
        """
        Return the stopped time during activity.

        Args:
            tolerance (float, optional): Maximal distance between two
                points for movement. Defaults to 2.45.

        Returns:
            datetime: Stopped time.
        """
        stopped_time = self.start_time() - self.start_time()  # Better way to do it?

        previous_point = self.trk[0].trkseg[0].trkpt[0]

        for track in self.trk:
            for segment in track.trkseg:
                for point in segment.trkpt:
                    if haversine_distance(previous_point, point) < tolerance:
                        stopped_time += point.time - previous_point.time
                    previous_point = point

        return stopped_time

    def moving_time(self) -> datetime:
        """
        Return the moving time.
        """
        return self.total_elapsed_time() - self.stopped_time()

    ###############################################################################
    #### Speed and Pace ###########################################################
    ###############################################################################

    def avg_speed(self) -> float:
        """
        Return the average speed (in kilometers per hour).
        """
        total_elapsed_time = (
            self.total_elapsed_time().total_seconds() / 3600
        )  # Total elapsed time in hours
        distance = self.distance() / 1000  # Distance in kilometers
        return distance / total_elapsed_time

    def avg_moving_speed(self) -> float:
        """
        Return the average moving speed (in kilometers per hour).
        """
        moving_time = self.moving_time().total_seconds() / 3600  # Moving time in hours
        distance = self.distance() / 1000  # Distance in kilometers
        return distance / moving_time

    def _compute_speed(self) -> None:
        """
        Compute speed (in kilometers per hour) at each track point.
        """
        if self._speed:
            return
        previous_point = self.first_point()
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    distance = (
                        haversine_distance(previous_point, track_point) / 1000
                    )  # Distance in kilometers
                    time = (
                        track_point.time - previous_point.time
                    ).total_seconds() / 3600  # Time in hours
                    try:
                        track_point.speed = distance / time
                    except ZeroDivisionError:
                        track_point.speed = 0.0
                    previous_point = track_point
        self._speed = True

    def min_speed(self) -> float:
        """
        Return the minimum speed.
        """
        self._compute_speed()
        min_speed = 1_000.0
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    min_speed = min(min_speed, track_point.speed)
        return min_speed

    def max_speed(self) -> float:
        """
        Return the maximum speed.
        """
        self._compute_speed()
        max_speed = -1.0
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    max_speed = max(max_speed, track_point.speed)
        return max_speed

    def avg_pace(self) -> float:
        """
        Return the average pace (in minute per kilometer).
        """
        return 60.0 / self.avg_speed()

    def avg_moving_pace(self) -> float:
        """
        Return the average moving pace (in minute per kilometer).
        """
        return 60.0 / self.avg_moving_speed()

    def _compute_pace(self) -> None:
        """
        Compute pace (in minutes per kilometer) at each track point.
        """
        if self._pace:
            return
        self._compute_speed()
        for track in self.trk:
            for segment in track.trkseg:
                for point in segment.trkpt:
                    try:
                        point.pace = 60.0 / point.speed
                    except ZeroDivisionError:
                        # Fill with average moving pace (first point)
                        point.pace = self.avg_moving_pace()
        self._pace = True

    def min_pace(self) -> float:
        """
        Return the minimum pace (in minutes per kilometer).
        """
        self._compute_pace()
        min_pace = 1000.0
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    min_pace = min(min_pace, track_point.pace)
        return min_pace

    def max_pace(self) -> float:
        """
        Return the maximum pace (in minutes per kilometer).
        """
        self._compute_pace()
        max_pace = -1.0
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    max_pace = max(max_pace, track_point.pace)
        return max_pace

    def _compute_ascent_speed(self) -> None:
        """
        Compute ascent speed (in kilometers per hour) at each track point.
        """
        if self._ascent_speed:
            return
        previous_point = self.first_point()
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    ascent = track_point.ele - previous_point.ele
                    # Convert to hours
                    time = (
                        track_point.time - previous_point.time
                    ).total_seconds() / 3600
                    try:
                        track_point.ascent_speed = ascent / time
                    except ZeroDivisionError:
                        track_point.ascent_speed = 0.0
                    previous_point = track_point
        self._ascent_speed = True

    def min_ascent_speed(self) -> float:
        """
        Return the minimum ascent speed (in meters per hour).
        """
        self._compute_ascent_speed()
        min_ascent_speed = 1000.0
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    min_ascent_speed = min(min_ascent_speed, track_point.ascent_speed)
        return min_ascent_speed

    def max_ascent_speed(self) -> float:
        """
        Return the maximum ascent speed (in meters per hour).
        """
        self._compute_ascent_speed()
        max_ascent_speed = -1.0
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    max_ascent_speed = max(max_ascent_speed, track_point.ascent_speed)
        return max_ascent_speed

    ###############################################################################
    #### Data Removal #############################################################
    ###############################################################################

    def remove_metadata(self):
        """
        Remove metadata.
        """
        self.metadata = None

    def remove_elevation(self):
        """
        Remove elevation data.
        """
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    track_point.ele = None

    def remove_time(self):
        """
        Remove time data.
        """
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    track_point.time = None

    def remove_extensions(self):
        """
        Remove extensions data.
        """
        # Remove extensions from gpx
        self.extensions = None

        # Remove extensions from metadata
        if self.metadata is not None:
            self.metadata.extensions = None

        # Remove extensions from waypoints
        if self.wpt is not None:
            for pt in self.wpt:
                pt.extensions = None

        # Remove extensions from routes
        if self.rte is not None:
            for rt in self.rte:
                rt.extensions = None

        # Remove extensions from tracks
        if self.trk is not None:
            for track in self.trk:
                track.extensions = None
                # Remove extensions from track segments
                if track.trkseg is not None:
                    for track_segment in track.trkseg:
                        track_segment.extensions = None
                        if track_segment.trkpt is not None:
                            for track_point in track_segment.trkpt:
                                # Remove extensions from track points
                                track_point.extensions = None

    ###############################################################################
    #### Error Correction #########################################################
    ###############################################################################

    def remove_points(self, remove_factor: int = 2):
        """
        TODO

        Args:
            remove_factor (int, optional): _description_. Defaults to 2.
        """
        count = 0
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if count % remove_factor == 0:
                        track_segment.trkpt.remove(track_point)
                        count += 1

    def remove_gps_errors(self, error_distance: float = 100) -> list:
        """
        Remove GPS errors.

        Args:
            error_distance (float, optional): Error threshold distance
                (in meters) between two points. Defaults to 100.

        Returns:
            list: List of removed points (GPS errors).
        """
        previous_point = None
        gps_errors = []

        for track in self.trk:
            for track_segment in track.trkseg:
                new_trkpt = []

                for track_point in track_segment.trkpt:
                    # GPS error
                    dst = haversine_distance(previous_point, track_point)
                    if (
                        previous_point is not None and dst > error_distance
                    ):  # TODO use Z-score?
                        gps_errors.append(track_point)
                    # No GPS error
                    else:
                        new_trkpt.append(track_point)
                        previous_point = track_point

                track_segment.trkpt = new_trkpt

        return gps_errors

    def remove_close_points(self, min_dist: float = 1, max_dist: float = 10):
        """
        Remove points that are to close together.

        Args:
            min_dist (float, optional): Minimal distance (in meters)
                between two points. Defaults to 1.
            max_dist (float, optional): Maximal distance (in meters)
                between two points. Defaults to 10.
        """
        point_1 = None
        point_2 = None

        for track in self.trk:
            for segment in track.trkseg:
                new_trkpt = []

                for point in segment.trkpt:
                    if point_1 is None:
                        point_1 = point
                        new_trkpt.append(point_1)
                    elif point_2 is None:
                        point_2 = point
                    else:
                        if (
                            haversine_distance(point_1, point_2) < min_dist
                            or haversine_distance(point_2, point) < min_dist
                        ) and haversine_distance(point_1, point) < max_dist:
                            point_2 = point
                        else:
                            new_trkpt.append(point_2)
                            point_1 = point_2
                            point_2 = point

                segment.trkpt = new_trkpt

    ###############################################################################
    #### Simplification ###########################################################
    ###############################################################################

    def simplify(self, epsilon: float):
        """
        Simplify GPX trk using Ramer-Douglas-Peucker algorithm.

        Args:
            epsilon (float): Tolerance.  TODO: add default value
        """
        for track in self.trk:
            for segment in track.trkseg:
                segment.trkpt = ramer_douglas_peucker(segment.trkpt, epsilon)

    ###############################################################################
    #### Exports ##################################################################
    ###############################################################################

    def to_dict(
        self,
        values: list[str] = None,
        as_series: bool = False,
    ) -> dict:
        """
        Convert GPX object to dictionary (similar to Polars
        `to_dict`).

        Args:
            values (list[str], optional): List of values to write.
                Supported values: "lat", "lon", "ele", "time", "speed",
                "pace", "ascent_rate", "ascent_speed",
                "distance_from_start". Defaults to None.
            as_series (bool, optional): True -> Values are Series,
                False -> Values are list[Any]. Defaults to False.

        Returns:
            dict: Dictionary representing the GPX.
        """
        return self.to_polars(values).to_dict(as_series=as_series)

    def to_dicts(
        self,
        values: list[str] = None,
    ) -> list[dict[str, Any]]:
        """
        Convert GPX object to list of dictionaries (similar to Polars
        `to_dicts`).

        Args:
            values (list[str], optional): List of values to write.
                Supported values: "lat", "lon", "ele", "time", "speed",
                "pace", "ascent_rate", "ascent_speed",
                "distance_from_start". Defaults to None.

        Returns:
            list[dict[str, Any]]: List of dictionaries representing the GPX.
        """
        return self.to_polars(values).to_dicts()

    def _to_dict_df(self, values: list[str] = None) -> dict:
        """
        Convert GPX object to dictionary.

        Args:
            values (list[str], optional): List of values to write.
                Supported values: "lat", "lon", "ele", "time", "speed",
                "pace", "ascent_rate", "ascent_speed",
                "distance_from_start". Defaults to None.

        Returns:
            dict: Dictionary containing data from GPX.
        """
        # Set default parameter
        if values is None:
            values = ["lat", "lon"]

        # Compute required values
        test_point = self.first_point()
        if "speed" in values and test_point.speed is None:
            self._compute_speed()
        if "pace" in values and test_point.pace is None:
            self._compute_pace()
        if "ascent_rate" in values and test_point.ascent_rate is None:
            self._compute_ascent_rate()
        if "ascent_speed" in values and test_point.ascent_speed is None:
            self._compute_ascent_speed()
        if "distance_from_start" in values and test_point.distance_from_start is None:
            self._compute_distance_from_start()

        # Create dataframe
        gpx_data = {}
        for v in values:
            if v == "time":
                gpx_data[v] = [
                    str(trkpt.time.replace(tzinfo=timezone.utc).astimezone(tz=None))
                    for trk in self.trk
                    for trkseg in trk.trkseg
                    for trkpt in trkseg.trkpt
                ]
            else:
                gpx_data[v] = [
                    getattr(trkpt, v)
                    for trk in self.trk
                    for trkseg in trk.trkseg
                    for trkpt in trkseg.trkpt
                ]
        return gpx_data

    def to_pandas(self, values: list[str] = None) -> pd.DataFrame:
        """
        Convert GPX object to Pandas Dataframe.
        Missing values are filled with default values (0 for numerical
        values and empty string for text).

        Args:
            values (list[str], optional): List of values to write.
                Supported values: "lat", "lon", "ele", "time", "speed",
                "pace", "ascent_rate", "ascent_speed",
                "distance_from_start". Defaults to None.

        Returns:
            pd.DataFrame: Dataframe containing data from GPX.
        """
        return pd.DataFrame(self._to_dict_df(values))

    def to_polars(self, values: list[str] = None) -> pl.DataFrame:
        """
        Convert GPX object to Polars Dataframe.
        Missing values are filled with default values (0 for numerical
        values and empty string for text).

        Args:
            values (list[str], optional): List of values to write.
                Supported values: "lat", "lon", "ele", "time", "speed",
                "pace", "ascent_rate", "ascent_speed",
                "distance_from_start". Defaults to None.

        Returns:
            pl.DataFrame: Dataframe containing data from GPX.
        """
        return pl.DataFrame(self._to_dict_df(values))

    def to_csv(
        self,
        dest: Optional[str | Path | IO[str] | IO[bytes] | bytes] = None,
        values: list[str] = None,
        **kwargs,
    ) -> str | None:
        """
        Write the GPX object track coordinates to a CSV file.

        Args:
            dest (str | Path | IO[str] | IO[bytes] | bytes, optional):
                Path to a file or a file-like object to write in.
                Defaults to None.
            values (list[str], optional): list of values to write.
                Supported values: "lat", "lon", "ele", "time", "speed",
                "pace", "ascent_rate", "ascent_speed",
                "distance_from_start". Defaults to None.

        Returns:
            str | None: CSV like string if path is set to None.
        """
        if values is None:
            values = ["lat", "lon"]

        if isinstance(dest, bytes):
            dest = io.BytesIO(dest)

        # Argument columns is required for KML writer (keep values order)
        return self.to_polars(values).select(values).write_csv(dest, **kwargs)



"""
This module contains functions to validate XML files against XML schemas.
"""

try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

import warnings
from pathlib import Path
from typing import Optional

import xmlschema


def check_xml_schema(file_path: str | Path, version: Optional[str] = None) -> bool:
    """
    Check XML schema.

    Parameters
    ----------
    file_path : str
        File path

    Returns
    -------
    bool
        True if the file follows XML schemas
    """
    file_path = Path(file_path)
    schema = None

    # GPX
    if file_path.suffix == ".gpx":
        if version is None or version == "1.1":
            schema = xmlschema.XMLSchema(
                files("ezgpx.schemas").joinpath("gpx_1_1/gpx.xsd")
            )
        elif version == "1.0":
            schema = xmlschema.XMLSchema(
                files("ezgpx.schemas").joinpath("gpx_1_0/gpx.xsd")
            )
        else:
            warnings.warn("Unable to check XML schema (unsupported GPX version)")
            return False

    # KML
    elif file_path.suffix == ".kml":
        schema = xmlschema.XMLSchema(
            files("ezgpx.schemas").joinpath("kml_2_2/ogckml22.xsd")
        )

    # KMZ
    elif file_path.suffix == ".kmz":
        return False

    # FIT
    elif file_path.suffix == ".fit":
        warnings.warn("Unable to check XML schema (fit files are not XML files)")
        return False

    # NOT SUPPORTED
    else:
        warnings.warn("Unable to check XML schema (unable to identify file type)")
        return False

    if schema is not None:
        return schema.is_valid(file_path)
    warnings.warn("Unable to check XML schema (unable to load XML schema)")
    return False


def check_xml_extensions_schemas(file_path: str | Path) -> bool:
    """
    Check XML extensions schemas.

    Parameters
    ----------
    file_path : str
        Path to GPX file.

    Returns
    -------
    bool
        True if GPX file follows XML extensions schemas.
    """
    raise NotImplementedError(
        "XML extensions schema validation is not implemented yet."
    )
    # gpx_schemas = [s for s in self.xsi_schema_location if s.endswith(".xsd")]
    # gpx_schemas.remove("http://www.topografix.com/GPX/1/1/gpx.xsd")
    # for gpx_schema in gpx_schemas:
    #     schema = xmlschema.XMLSchema(gpx_schema)
    #     if not schema.is_valid(file_path):
    #         warnings.warn(f"File does not follow {gpx_schema}")
    #         return False
    # return True



import math
import os
import sys

import pytest

file_folder = os.path.dirname(__file__)
parent_folder = os.path.realpath(os.path.dirname(file_folder))  # ie: ezGPX

os.chdir(file_folder)
sys.path.append(parent_folder + "/ezgpx")

from ezgpx import WayPoint, utils


class TestUtils:
    def test_haversine_distance(self):
        point_1 = WayPoint("wpt", 48.0, 2.0)
        point_2 = WayPoint("wpt", 43.0, 5.0)
        assert utils.haversine_distance(point_1, point_2) == pytest.approx(603020.0)

    @pytest.mark.parametrize("start,end,point,expected", [
        pytest.param(WayPoint("wpt", 0, 0), WayPoint("wpt", 0, 2), WayPoint("wpt", 1, 1), 1.0, id="horizontal_line"),
        pytest.param(WayPoint("wpt", 0, 0), WayPoint("wpt", 2, 0), WayPoint("wpt", 1, 1), 1.0, id="vertical_line"),
        pytest.param(WayPoint("wpt", 0, 0), WayPoint("wpt", 1, 1), WayPoint("wpt", 1, 0), math.sqrt(2) / 2, id="diagonal_line"),
        pytest.param(WayPoint("wpt", 0, 0), WayPoint("wpt", 1, 1), WayPoint("wpt", 2, 2), 0.0, id="point_on_line"),
    ])
    def test_perpendicular_distance(self, start, end, point, expected):
        assert utils.perpendicular_distance(start, end, point) == pytest.approx(expected)



"""
This module contains the Metadata class.
"""

from datetime import datetime

from .extensions import Extensions
from .gpx_element import GpxElement
from .link import Link
from .person import Person


class Metadata(GpxElement):
    """
    metadataType element in GPX file.
    """

    fields = [
        "name",
        "desc",
        "author",
        "copyright",
        "link",
        "time",
        "keywords",
        "bounds",
        "extensions",
    ]
    mandatory_fields = []

    def __init__(
        self,
        tag: str = "metadata",
        name: str = None,
        desc: str = None,
        author: Person = None,
        copyright: str = None,  # pylint: disable=redefined-builtin
        link: Link = None,
        time: datetime = None,
        keywords: str = None,
        bounds: str = None,
        extensions: Extensions = None,
    ) -> None:
        """
        Initialise Metadata instance.

        Args:
            tag (str, optional): XML tag. Defaults to "metadata".
            name (str, optional): Name. Defaults to None.
            desc (str, optional): Description. Defaults to None.
            author (Person, optional): Author. Defaults to None.
            copyright (str, optional): Copyright. Defaults to None.
            link (Link, optional): Link. Defaults to None.
            time (datetime, optional): Time. Defaults to None.
            keywords (str, optional): Keywords. Defaults to None.
            bounds (str, optional): Bounds. Defaults to None.
            extensions (Extensions, optional): Extensions. Defaults to None.
        """
        self.tag: str = tag
        self.name: str = name
        self.desc: str = desc
        self.author: Person = author
        self.copyright: str = copyright
        self.link: Link = link
        self.time: datetime = time
        self.keywords: str = keywords
        self.bounds: str = bounds
        self.extensions: str = extensions



"""
This file contains the high level GPX object.
"""

from __future__ import annotations

import io
import warnings
from datetime import datetime
from math import degrees
from pathlib import Path
from typing import IO, Optional
from zipfile import ZipFile

import pandas as pd
import polars as pl
from narwhals.typing import IntoFrameT

from ..constants.precisions import DEFAULT_PRECISION_DICT, DEFAULT_TIME_FORMAT
from ..gpx_elements import (
    Bounds,
    Copyright,
    Email,
    Extensions,
    Gpx,
    Link,
    Metadata,
    Person,
    Point,
    PointSegment,
    Route,
    Track,
    TrackSegment,
    WayPoint,
)
from ..parsers.fit_parser import FitParser
from ..parsers.gpx_parser import GPXParser
from ..parsers.kml_parser import KMLParser
from ..utils import EARTH_RADIUS, check_xml_extensions_schemas, check_xml_schema
from ..utils.dataframe import is_dataframe
from ..writers.gpx_writer import GPXWriter
from ..writers.kml_writer import KMLWriter


class GPX:
    """
    High level GPX object.
    """

    TIME_RELATED_VALUES = ["time", "speed", "pace", "ascent_speed"]
    ELEVATION_RELATED_VALUES = ["ele", "ascent_rate", "ascent_speed"]

    def __init__(
        self,
        source: Optional[str | Path | IO[str] | IO[bytes] | bytes | IntoFrameT] = None,
        xml_schema: bool = True,
        xml_extensions_schemas: bool = False,
    ) -> None:
        """
        Initialise high level GPX object.

        Args:
            source (str | Path | IO[str] | IO[bytes] | bytes | IntoFrameT, optional):
                Path to a file or a file-like object to parse. Defaults to None.
            xml_schemas (bool, optional): Toggle schema
                verification during parsing. Defaults to True.
            xml_extensions_schemas (bool, optional): Toggle extensions
                schema verificaton durign parsing. Requires internet connection
                connection and is not guaranted to work. Defaults to False.
        Raises:
            FileNotFoundError: _description_
            ValueError: _description_
            FileNotFoundError: _description_
        """
        # GPX file description
        self.source: str | Path | IO[str] | IO[bytes] | bytes | IntoFrameT = None

        # GPX file content
        self.gpx: Gpx = None
        self._ele_data: bool = False
        self._time_data: bool = False
        self._precisions: dict = DEFAULT_PRECISION_DICT
        self._time_format: str = DEFAULT_TIME_FORMAT
        self._extensions_fields: dict = {}

        # Writers
        self._gpx_writer: GPXWriter = None
        self._kml_writer: KMLWriter = None

        # Empty source - Create empty GPX instance for advanced use only
        if source is None:
            self._init_from_none()
        # Valid file
        elif isinstance(source, (str, Path)):
            self.source = Path(source)

            # GPX
            if self.source.suffix == ".gpx":
                self._init_from_gpx(xml_schema, xml_extensions_schemas)

            # KML
            elif self.source.suffix == ".kml":
                self._init_from_kml()

            # KMZ
            elif self.source.suffix == ".kmz":
                self._init_from_kmz(xml_schema, xml_extensions_schemas)

            # FIT
            elif self.source.suffix == ".fit":
                self._init_from_fit()

            # Invalid file or file path
            else:
                raise ValueError(
                    f"Unable to parse this type of file: {source}"
                    "Consider renaming your file with the proper file extension."
                )

            # Writers
            self._gpx_writer = GPXWriter(self.gpx, self._precisions, self._time_format)
            self._kml_writer = KMLWriter(
                self.gpx, precisions=self._precisions, time_format=self._time_format
            )

        # Dataframe
        elif is_dataframe(source):
            self._init_from_dataframe(source)
        else:
            raise TypeError(
                f"Invalid source type: {type(source)}. "
                "Expected str, Path, IO[str], IO[bytes], bytes, pd.DataFrame or pl.DataFrame."
            )

    def _init_from_none(self):
        """
        Initialise empty GPX instance.
        """
        warnings.warn(
            "No file path provided, creating an empty GPX instance.", UserWarning
        )
        self.gpx = Gpx()

        # Writers
        self._gpx_writer = GPXWriter(self.gpx)
        self._kml_writer = KMLWriter(self.gpx)

    def _init_from_gpx(
        self, xml_schema: bool = True, xml_extensions_schemas: bool = False
    ):
        """
        Initialise GPX instance from GPX file.

        Args:
            xml_schemas (bool, optional): Toggle schema
                verification during parsing. Defaults to True.
            xml_extensions_schemas (bool, optional): Toggle extensions
                schema verificaton durign parsing. Requires internet connection
                connection and is not guaranted to work. Defaults to False.
        """
        parser = GPXParser(self.source, xml_schema, xml_extensions_schemas)
        self.gpx = parser.gpx
        self._ele_data = parser.ele_data
        self._time_data = parser.time_data
        self._precisions = parser.precisions
        self._time_format = parser.time_format
        self._extensions_fields = parser.extensions_fields

    def _init_from_kml(
        self, xml_schema: bool = True, xml_extensions_schemas: bool = False
    ):
        """
        Initialise GPX instance from KML file.

        Args:
            xml_schemas (bool, optional): Toggle schema
                verification during parsing. Defaults to True.
            xml_extensions_schemas (bool, optional): Toggle extensions
                schema verificaton durign parsing. Requires internet connection
                connection and is not guaranted to work. Defaults to False.
        """
        parser = KMLParser(self.source, xml_schema, xml_extensions_schemas)
        self.gpx = parser.gpx
        self._precisions = parser.precisions
        self._time_format = parser.time_format

    def _init_from_kmz(
        self, xml_schema: bool = True, xml_extensions_schemas: bool = False
    ):
        """
        Initialise GPX instance from KMZ file.

        Args:
            xml_schemas (bool, optional): Toggle schema
                verification during parsing. Defaults to True.
            xml_extensions_schemas (bool, optional): Toggle extensions
                schema verificaton durign parsing. Requires internet connection
                connection and is not guaranted to work. Defaults to False.
        """
        with ZipFile(self.source, "r") as kmz:
            kmls = [
                info.filename
                for info in kmz.infolist()
                if info.filename.endswith(".kml")
            ]
            if "doc.kml" not in kmls:
                raise FileNotFoundError(
                    f"Unable to parse file: {self.source}"
                    "Expected to find doc.kml inside KMZ file."
                )
            with kmz.open("doc.kml", "r") as kml_file:
                kml = io.BytesIO(kml_file.read())
        parser = KMLParser(kml, xml_schema, xml_extensions_schemas)
        self.gpx = parser.gpx
        self._precisions = parser.precisions
        self._time_format = parser.time_format

    def _init_from_fit(self):
        """
        Initialise GPX instance from FIT file.
        """
        parser = FitParser(self.source)
        self.gpx = parser.gpx
        self._precisions = parser.precisions
        self._time_format = parser.time_format

    def _init_from_dataframe(self, source: IntoFrameT):
        """
        Initialise GPX instance from Dataframe.
        """
        self.gpx = Gpx.from_dataframe(source)

    def __str__(self) -> str:
        return self._gpx_writer.gpx_to_string()

    def __repr__(self):
        return f"source = {self.source}\ngpx = {self.gpx}"

    ###############################################################################
    #### Schemas ##################################################################
    ###############################################################################

    def check_xml_schema(self) -> bool:
        """
        Check XML schema.

        Returns:
            bool: True if the file follows XML schemas.
        """
        return check_xml_schema(self.source, self.gpx.version)

    def check_xml_extensions_schemas(self) -> bool:
        """
        Check XML extension schemas.

        Returns:
            bool: True if the file follows XML schemas.
        """
        return check_xml_extensions_schemas(self.source)

    ###############################################################################
    #### Name #####################################################################
    ###############################################################################

    def name(self) -> str:
        """
        Return activity name.
        """
        return self.gpx.name()

    def set_name(self, new_name: str) -> None:
        """
        Set activity name.

        Args:
            new_name (str): New activity name.
        """
        self.gpx.set_name(new_name)

    ###############################################################################
    #### Points ###################################################################
    ###############################################################################

    def nb_points(self) -> int:
        """
        Return the number of points in the GPX.
        """
        return self.gpx.nb_points()

    def bounds(self) -> tuple[float, float, float, float]:
        """
        Find the bounding coordinates of the activity.

        Returns:
            tuple[float, float, float, float]: Min latitude, min
                longitude, max latitude, max longitude.
        """
        return self.gpx.bounds()

    def center(self) -> tuple[float, float]:
        """
        Find the center coordinates of the activity.

        Returns:
            tuple[float, float]: Latitude and longitude of the center
                point.
        """
        return self.gpx.center()

    def first_point(self) -> WayPoint:  # TODO remove?
        """
        Return the first point of the activity.
        """
        return self.gpx.first_point()

    def last_point(self) -> WayPoint:  # TODO remove?
        """
        Return the last point of the activity.
        """
        return self.gpx.last_point()

    def extreme_points(
        self,
    ) -> tuple[WayPoint, WayPoint, WayPoint, WayPoint]:  # TODO remove?
        """
        Find extreme points in track, i.e.: points with lowest and
        highest latitude and longitude.

        Returns:
            tuple[WayPoint, WayPoint, WayPoint, WayPoint]: Min latitude
                point, min longitude point, max latitude point, max
                longitude points.
        """
        return self.gpx.extreme_points()

    ###############################################################################
    #### Distance and Elevation ###################################################
    ###############################################################################

    def distance(self) -> float:
        """
        Returns the distance (in meters) of the activity.
        """
        return self.gpx.distance()

    def ascent(self) -> float:
        """
        Returns the ascent (in meters) of the activity.
        """
        return self.gpx.ascent()

    def descent(self) -> float:
        """
        Returns the descent (in meters) of the activity.
        """
        return self.gpx.descent()

    def min_elevation(self) -> float:
        """
        Returns the minimum elevation (in meters) of the activity.
        """
        return self.gpx.min_elevation()

    def max_elevation(self) -> float:
        """
        Returns the maximum elevation (in meters) of the activity.
        """
        return self.gpx.max_elevation()

    def max_descent_rate(self) -> float:
        """
        Return the minimum ascent rate of the activity.
        """
        return self.gpx.max_descent_rate()

    def max_ascent_rate(self) -> float:
        """
        Return the maximum ascent rate of the activity.
        """
        return self.gpx.max_ascent_rate()

    ###############################################################################
    #### Time #####################################################################
    ###############################################################################

    def start_time(self) -> datetime:
        """
        Return the start time of the activity.
        """
        return self.gpx.start_time()

    def stop_time(self) -> datetime:
        """
        Return the stop time of the activity.
        """
        return self.gpx.stop_time()

    def total_elapsed_time(self) -> datetime:
        """
        Return the total elapsed time.
        """
        return self.gpx.total_elapsed_time()

    def stopped_time(self) -> datetime:
        """
        Return the stopped time.
        """
        return self.gpx.stopped_time()

    def moving_time(self) -> datetime:
        """
        Return the moving time.
        """
        return self.gpx.moving_time()

    ###############################################################################
    #### Speed and Pace ###########################################################
    ###############################################################################

    def avg_speed(self) -> float:
        """
        Return average speed (in kilometers per hour) of the activity.
        """
        return self.gpx.avg_speed()

    def avg_moving_speed(self) -> float:
        """
        Return average moving speed (in kilometers per hour).
        """
        return self.gpx.avg_moving_speed()

    def min_speed(self) -> float:
        """
        Return the minimum speed (in kilometers per hour)during the activity.
        """
        return self.gpx.min_speed()

    def max_speed(self) -> float:
        """
        Return the maximum speed (in kilometers per hour).
        """
        return self.gpx.max_speed()

    def avg_pace(self) -> float:
        """
        Return average pace (in minutes per kilometer).
        """
        return self.gpx.avg_pace()

    def avg_moving_pace(self) -> float:
        """
        Return average moving pace (in minutes per kilometer).
        """
        return self.gpx.avg_moving_pace()

    def min_pace(self) -> float:
        """
        Return the minimum pace (in minutes per kilometer).
        """
        return self.gpx.min_pace()

    def max_pace(self) -> float:
        """
        Return the maximum pace (in minutes per kilometer).
        """
        return self.gpx.max_pace()

    def min_ascent_speed(self) -> float:
        """
        Return the minimum ascent speed (in meters per hour).
        """
        return self.gpx.min_ascent_speed()

    def max_ascent_speed(self) -> float:
        """
        Return the maximum ascent speed (in meters per hour).
        """
        return self.gpx.max_ascent_speed()

    ###############################################################################
    #### Data Removal #############################################################
    ###############################################################################

    def remove_metadata(self):
        """
        Remove metadata.
        """
        self.gpx.remove_metadata()

    def remove_elevation(self):
        """
        Remove elevation data.
        """
        self.gpx.remove_elevation()

    def remove_time(self):
        """
        Remove time data.
        """
        self.gpx.remove_time()

    def remove_extensions(self):
        """
        Remove extensions data.
        """
        self.gpx.remove_extensions()

    ###############################################################################
    #### Error Correction #########################################################
    ###############################################################################

    def remove_gps_errors(self):
        """
        Remove GPS errors.
        """
        self.gpx.remove_gps_errors()

    def remove_close_points(self, min_dist: float = 1, max_dist: float = 10):
        """
        Remove points that are to close together.

        Args:
            min_dist (float, optional): Minimal distance between two
                points. Defaults to 1.
            max_dist (float, optional): Maximal distance between two
                points. Defaults to 10.
        """
        self.gpx.remove_close_points(min_dist, max_dist)

    ###############################################################################
    #### Simplification ###########################################################
    ###############################################################################

    def simplify(self, tolerance: float = 2):
        """
        Simplify tracks using Ramer-Douglas-Peucker algorithm.

        Args:
            tolerance (float, optional): Minimum distance (in meters)
            between the point and the track before the point is
            removed. Defaults to 2.
        """
        epsilon = degrees(tolerance / EARTH_RADIUS)
        self.gpx.simplify(epsilon)

    ###############################################################################
    #### Merge ####################################################################
    ###############################################################################

    @staticmethod
    def merge(gpx_1: GPX, gpx_2: GPX) -> GPX:
        """
        Merge GPX objects into a new instance.

        Args:
            gpx_1 (GPX): First GPX object.
            gpx_2 (GPX): Second GPX object.

        Returns:
            GPX: Merged GPX (new instance).
        """
        topo = [
            "http://www.topografix.com/GPX/1/1",
            "http://www.topografix.com/GPX/1/1/gpx.xsd",
        ]

        # Create new GPX instance
        merged_gpx = GPX()

        # Fill new GPX instance
        merged_gpx.gpx.tag = "gpx"
        merged_gpx.gpx.xmlns = "http://www.topografix.com/GPX/1/1"
        merged_gpx.gpx.xsi_schema_location = list(
            set(topo + gpx_1.gpx.xsi_schema_location + gpx_2.gpx.xsi_schema_location)
        )
        merged_gpx.gpx.version = "1.1"
        merged_gpx.gpx.creator = "ezGPX"
        merged_gpx.gpx.metadata = (
            gpx_2.gpx.metadata if gpx_1.gpx.metadata is None else gpx_1.gpx.metadata
        )
        merged_gpx.gpx.wpt = gpx_1.gpx.wpt + gpx_2.gpx.wpt
        merged_gpx.gpx.rte = gpx_1.gpx.rte + gpx_2.gpx.rte
        merged_gpx.gpx.trk = gpx_1.gpx.trk + gpx_2.gpx.trk
        merged_gpx.gpx.extensions = Extensions(
            "extensions", gpx_1.gpx.metadata | gpx_2.gpx.metadata
        )

        # Return new GPX instance
        return merged_gpx

    ###############################################################################
    #### Exports ##################################################################
    ###############################################################################

    def to_dict(
        self,
        values: list[str] = None,
        as_series: bool = False,
    ) -> dict:
        """
        Convert GPX object to dictionary.

        Args:
            values (list[str], optional): list of values to write.
                Supported values: "lat", "lon", "ele", "time", "speed",
                "pace", "ascent_rate", "ascent_speed",
                "distance_from_start". Defaults to None.
            as_series (bool, optional): True -> Values are Series False
                -> Values are list[Any]. Defaults to False.

        Returns:
            dict: Return a dictionary representing the GPX.
        """
        return self.gpx.to_dict(values, as_series)

    def to_pandas(self, values: list[str] = None) -> pd.DataFrame:
        """
        Convert GPX object to Pandas Dataframe.

        Args:
            values (list[str], optional): list of values to write.
                Supported values: "lat", "lon", "ele", "time", "speed",
                "pace", "ascent_rate", "ascent_speed",
                "distance_from_start". Defaults to None.

        Returns:
            pd.DataFrame: Return a dictionary representing the GPX.
        """
        # Set default parameter
        if values is None:
            values = ["lat", "lon"]

        # Disable time related values if no time data available
        if not self._time_data:
            if any(v in GPX.TIME_RELATED_VALUES for v in values):
                warnings.warn(
                    f"""Trying to create dataframe from GPX file {self.source}
                        which does not contain time data. Time related values
                        (time, speed, pace, ascent speed) will not be present in
                        the dataframe.""",
                    UserWarning,
                )
            for v in GPX.TIME_RELATED_VALUES:
                if v in values:
                    values.remove(v)

        # Disable elevation related values if no elevation data available
        if not self._ele_data:
            if any(v in GPX.ELEVATION_RELATED_VALUES for v in values):
                warnings.warn(
                    f"""Trying to create dataframe from GPX file {self.source}
                        which does not contain elevation data. Time related
                        values (elevation, ascent rate, ascent speed) will not
                        be present in the dataframe.""",
                    UserWarning,
                )
            for v in GPX.ELEVATION_RELATED_VALUES:
                if v in values:
                    values.remove(v)

        return self.gpx.to_pandas(values)

    def to_polars(self, values: list[str] = None) -> pl.DataFrame:
        """
        Convert GPX object to Polars Dataframe.

        Args:
            values (list[str], optional): list of values to write.
                Supported values: "lat", "lon", "ele", "time", "speed",
                "pace", "ascent_rate", "ascent_speed",
                "distance_from_start". Defaults to None.

        Returns:
            pl.DataFrame: Return a dictionary representing the GPX.
        """
        # Set default parameter
        if values is None:
            values = ["lat", "lon"]

        # Disable time related values if no time data available
        if not self._time_data:
            if any(v in GPX.TIME_RELATED_VALUES for v in values):
                warnings.warn(
                    f"""Trying to create dataframe from GPX file {self.source}
                        which does not contain time data. Time related values
                        (time, speed, pace, ascent speed) will not be present in
                        the dataframe.""",
                    UserWarning,
                )
            for v in GPX.TIME_RELATED_VALUES:
                if v in values:
                    values.remove(v)

        # Disable elevation related values if no elevation data available
        if not self._ele_data:
            if any(v in GPX.ELEVATION_RELATED_VALUES for v in values):
                warnings.warn(
                    f"""Trying to create dataframe from GPX file {self.source}
                        which does not contain elevation data. Time related
                        values (elevation, ascent rate, ascent speed) will not
                        be present in the dataframe.""",
                    UserWarning,
                )
            for v in GPX.ELEVATION_RELATED_VALUES:
                if v in values:
                    values.remove(v)

        return self.gpx.to_polars(values)

    def to_csv(
        self,
        dest: Optional[str | Path | IO[str] | IO[bytes] | bytes] = None,
        values: list[str] = None,
        **kwargs,
    ) -> str | None:
        """
        Write the GPX object track coordinates to a CSV file.

        Args:
            dest (str | Path | IO[str] | IO[bytes] | bytes, optional):
                Path to a file or a file-like object to write in.
                Defaults to None.
            values (list[str], optional): list of values to write.
                Supported values: "lat", "lon", "ele", "time", "speed",
                "pace", "ascent_rate", "ascent_speed",
                "distance_from_start". Defaults to None.

        Returns:
            str | None: CSV like string if path is set to None.
        """
        return self.gpx.to_csv(dest, values, **kwargs)

    def to_gpx(
        self,
        dest: Optional[str | Path | IO[str] | IO[bytes] | bytes] = None,
        *,
        properties: bool = True,
        bounds_fields: Optional[list[str]] = None,
        copyright_fields: Optional[list[str]] = None,
        email_fields: Optional[list[str]] = None,
        extensions_fields: Optional[dict] = None,
        gpx_fields: Optional[list[str]] = None,
        link_fields: Optional[list[str]] = None,
        metadata_fields: Optional[list[str]] = None,
        person_fields: Optional[list[str]] = None,
        point_segment_fields: Optional[list[str]] = None,
        point_fields: Optional[list[str]] = None,
        route_fields: Optional[list[str]] = None,
        track_segment_fields: Optional[list[str]] = None,
        track_fields: Optional[list[str]] = None,
        waypoint_fields: Optional[list[str]] = None,
        track_point_fields: Optional[list[str]] = None,
        mandatory_fields: bool = True,
    ) -> str | None:
        """
        Write the GPX object to a GPX file.

        Args:
            dest (str | Path | IO[str] | IO[bytes] | bytes, optional):
                Path to a file or a file-like object to write in.
                Defaults to None.
            properties (bool, optional): _description_. Defaults to True.
            bounds_fields (Optional[list[str]], optional): _description_. Defaults to None.
            copyright_fields (Optional[list[str]], optional): _description_. Defaults to None.
            email_fields (Optional[list[str]], optional): _description_. Defaults to None.
            extensions_fields (Optional[dict], optional): _description_. Defaults to None.
            gpx_fields (Optional[list[str]], optional): _description_. Defaults to None.
            link_fields (Optional[list[str]], optional): _description_. Defaults to None.
            metadata_fields (Optional[list[str]], optional): _description_. Defaults to None.
            person_fields (Optional[list[str]], optional): _description_. Defaults to None.
            point_segment_fields (Optional[list[str]], optional): _description_. Defaults to None.
            point_fields (Optional[list[str]], optional): _description_. Defaults to None.
            route_fields (Optional[list[str]], optional): _description_. Defaults to None.
            track_segment_fields (Optional[list[str]], optional): _description_. Defaults to None.
            track_fields (Optional[list[str]], optional): _description_. Defaults to None.
            waypoint_fields (Optional[list[str]], optional): _description_. Defaults to None.
            track_point_fields (Optional[list[str]], optional): _description_. Defaults to None.
            mandatory_fields (bool, optional): _description_. Defaults to True.

        Returns:
            str | None: GPX like string if path is set to None.
        """
        bounds_fields = bounds_fields if bounds_fields is not None else Bounds.fields
        copyright_fields = (
            copyright_fields if copyright_fields is not None else Copyright.fields
        )
        email_fields = email_fields if email_fields is not None else Email.fields
        extensions_fields = (
            extensions_fields
            if extensions_fields is not None
            else self._extensions_fields
        )
        gpx_fields = gpx_fields if gpx_fields is not None else Gpx.fields
        link_fields = link_fields if link_fields is not None else Link.fields
        metadata_fields = (
            metadata_fields if metadata_fields is not None else Metadata.fields
        )
        person_fields = person_fields if person_fields is not None else Person.fields
        point_segment_fields = (
            point_segment_fields
            if point_segment_fields is not None
            else PointSegment.fields
        )
        point_fields = point_fields if point_fields is not None else Point.fields
        route_fields = route_fields if route_fields is not None else Route.fields
        track_segment_fields = (
            track_segment_fields
            if track_segment_fields is not None
            else TrackSegment.fields
        )
        track_fields = track_fields if track_fields is not None else Track.fields
        waypoint_fields = (
            waypoint_fields if waypoint_fields is not None else WayPoint.fields
        )
        track_point_fields = (
            track_point_fields if track_point_fields is not None else WayPoint.fields
        )
        return self._gpx_writer.write(
            file_path=dest,
            properties=properties,
            bounds_fields=bounds_fields,
            copyright_fields=copyright_fields,
            email_fields=email_fields,
            extensions_fields=extensions_fields,
            gpx_fields=gpx_fields,
            link_fields=link_fields,
            metadata_fields=metadata_fields,
            person_fields=person_fields,
            point_segment_fields=point_segment_fields,
            point_fields=point_fields,
            route_fields=route_fields,
            track_segment_fields=track_segment_fields,
            track_fields=track_fields,
            waypoint_fields=waypoint_fields,
            track_point_fields=track_point_fields,
            mandatory_fields=mandatory_fields,
        )

    def to_kml(
        self,
        dest: Optional[str | Path | IO[str] | IO[bytes] | bytes] = None,
        *,
        styles: Optional[list[tuple[str, dict]]] = None,
    ) -> str | None:
        """
        Write the GPX object to a KML file.

        Args:
            dest (str | Path | IO[str] | IO[bytes] | bytes, optional):
                Path to a file or a file-like object to write in.
                Defaults to None.
            styles (Optional[list[tuple[str, dict]]], optional): KML
                styles. Defaults to None.

        Returns:
            str | None: KML like string if path is set to None.
        """
        return self._kml_writer.write(dest, styles)



"""
This module contains the Bounds class.
"""

from .gpx_element import GpxElement


class Bounds(GpxElement):
    """
    boundsType element in GPX file.
    """

    fields = ["minlat", "minlon", "maxlat", "maxlon"]
    mandatory_fields = ["minlat", "minlon", "maxlat", "maxlon"]

    def __init__(
        self,
        tag: str = "bounds",
        minlat: float = None,
        minlon: float = None,
        maxlat: float = None,
        maxlon: float = None,
    ) -> None:
        """
        Initialise Bounds instance.

        Args:
            tag (str, optional): XML tag. Defaults to "bounds".
            minlat (float, optional): Minimal latitude (in degrees).
                Defaults to None.
            minlon (float, optional): Minimal longitude (in degrees).
                Defaults to None.
            maxlat (float, optional): Maximal latitude (in degrees).
                Defaults to None.
            maxlon (float, optional): Maximal longitude (in degrees).
                Defaults to None.
        """
        self.tag: str = tag
        self.minlat: float = minlat
        self.minlon: float = minlon
        self.maxlat: float = maxlat
        self.maxlon: float = maxlon



"""
This module contains utility functions.
"""

from typing import Any

import narwhals as nw


def is_dataframe(obj: Any) -> bool:
    """
    Check if an object is a dataframe.
    """
    try:
        nw.from_native(obj)
        return True
    except Exception:  # pylint: disable=broad-exception-caught
        return False



"""
This module contains the Route class.
"""

from .extensions import Extensions
from .gpx_element import GpxElement
from .link import Link
from .waypoint import WayPoint


class Route(GpxElement):
    """
    rteType element in GPX file.
    """

    fields = [
        "name",
        "cmt",
        "desc",
        "src",
        "link",
        "number",
        "type",
        "extensions",
        "rtept",
    ]
    mandatory_fields = []

    def __init__(
        self,
        tag: str = "rte",
        name: str = None,
        cmt: str = None,
        desc: str = None,
        src: str = None,
        link: Link = None,
        number: int = None,  # non negative integer
        type: str = None,  # pylint: disable=redefined-builtin
        extensions: Extensions = None,
        rtept: list[WayPoint] = None,
    ) -> None:
        """
        Initialise Route instance.

        Args:
            tag (str, optional): XML tag. Defaults to "rte".
            name (str, optional): Name. Defaults to None.
            cmt (str, optional): Comment. Defaults to None.
            desc (str, optional): Description. Defaults to None.
            src (str, optional): Source. Defaults to None.
            link (Link, optional): Link. Defaults to None.
            number (int, optional): Number. Defaults to None.
            type (str, optional): Type. Defaults to None.
            extensions (Extensions, optional): Extensions. Defaults to None.
            rtept (list[WayPoint], optional): List of route points.
                Defaults to None.
        """
        self.tag: str = tag
        self.name: str = name
        self.cmt: str = cmt
        self.desc: str = desc
        self.src: str = src
        self.link: Link = link
        self.number: int = number
        self.type: str = type
        self.extensions: Extensions = extensions
        self.rtept: list[WayPoint] = [] if rtept is None else rtept



"""
This module contains the WayPoint class.
"""

import datetime

from .extensions import Extensions
from .gpx_element import GpxElement
from .link import Link


class WayPoint(GpxElement):
    """
    wptType element in GPX file.
    """

    fields = [
        "lat",
        "lon",
        "ele",
        "time",
        "mag_var",
        "geo_id_height",
        "name",
        "cmt",
        "desc",
        "src",
        "link",
        "sym",
        "type",
        "fix",
        "sat",
        "hdop",
        "vdop",
        "pdop",
        "age_of_gps_data",
        "dgpsid",
        "extensions",
    ]
    mandatory_fields = ["lat", "lon"]

    def __init__(
        self,
        tag: str = "wpt",
        lat: float = None,
        lon: float = None,
        ele: float = None,
        time: datetime = None,
        mag_var: float = None,  # 0 <= value < 360
        geo_id_height: float = None,
        name: str = None,
        cmt: str = None,
        desc: str = None,
        src: str = None,
        link: Link = None,
        sym: str = None,
        type: str = None,  # pylint: disable=redefined-builtin
        fix: str = None,  # none, 2d, 3d, dgps, pps
        sat: int = None,  # non negative integer
        hdop: float = None,
        vdop: float = None,
        pdop: float = None,
        age_of_gps_data: float = None,
        dgpsid: int = None,  # 0 <= value <= 1023
        extensions: Extensions = None,
    ) -> None:
        """
        Initialise WayPoint instance.

        Args:
            tag (str, optional): XML tag. Defaults to "wpt".
            lat (float, optional): Latitude (in degrees). Defaults to None.
            lon (float, optional): Longitude (in degrees). Defaults to None.
            ele (float, optional): Elevation (in meters). Defaults to None.
            time (datetime, optional): Time. Defaults to None.
            mag_var (float, optional): Magnetic variation (in degrees)
                at the point. Defaults to None.
            geo_id_height (float, optional): Height (in meters) of
                geoid (mean sea level) above WGS84 earth ellipsoid.
                As defined in NMEA GGA message. Defaults to None.
            name (str, optional): Name. Defaults to None.
            cmt (str, optional): Comment. Defaults to None.
            desc (str, optional): Description. Defaults to None.
            src (str, optional): Source. Defaults to None.
            link (Link, optional): Link. Defaults to None.
            sym (str, optional): Text of GPS symbol name. For
                interchange with other programs, use the
                exact spelling of the symbol as displayed on the GPS.
                If the GPS abbreviates words, spell them out. Defaults
                to None.
            type (str, optional): Type. Defaults to None.
            fix (str, optional): Type of GPX fix. Defaults to None.
            sat (int, optional): Number of satellites used to calculate
                the GPX fix. Defaults to None.
            hdop (float, optional): Horizontal dilution of precision.
                Defaults to None.
            vdop (float, optional): Vertical dilution of precision.
                Defaults to None.
            pdop (float, optional): Position dilution of precision.
                Defaults to None.
            age_of_gps_data (float, optional): Number of seconds since
                last DGPS update. Defaults to None.
            dgpsid (int, optional): ID of DGPS station used in
                differential correction. Defaults to None.
            extensions (Extensions, optional): Extensions. Defaults to None.
        """
        self.tag: str = tag
        self.lat: float = lat
        self.lon: float = lon
        self.ele: float = ele
        self.time: datetime = time
        self.mag_var: float = mag_var
        self.geo_id_height: float = geo_id_height
        self.name: str = name
        self.cmt: str = cmt
        self.desc: str = desc
        self.src: str = src
        self.link: Link = link
        self.sym: str = sym
        self.type: str = type
        self.fix: str = fix
        self.sat: int = sat
        self.hdop: float = hdop
        self.vdop: float = vdop
        self.pdop: float = pdop
        self.age_of_gps_data: float = age_of_gps_data
        self.dgpsid: int = dgpsid
        self.extensions: Extensions = extensions

        # Statistics for map plotting
        # https://support.strava.com/hc/en-us/articles/360049869011-Personalized-Stat-Maps
        self.speed: float = None
        self.pace: float = None
        self.ascent_rate: float = None
        self.ascent_speed: float = None
        self.distance_from_start: float = None



"""
This module contains the GpxElement class.
"""


class GpxElement:
    """
    Base class for element in GPX file.
    Implements dunders functions.
    """

    fields = []
    mandatory_fields = []

    def __init__(self) -> None:
        pass

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}[{self.tag}]("
            + ", ".join(str(getattr(self, a)) for a in self.mandatory_fields)
            + ")"
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}[{self.tag}]("
            + ", ".join(repr(getattr(self, a)) for a in self.mandatory_fields)
            + ")"
        )



"""
This module contains the Parser class.
"""

import warnings
from datetime import datetime
from pathlib import Path
from typing import IO, Dict

from ..constants.precisions import (
    DEFAULT_PRECISION,
    DEFAULT_PRECISION_DICT,
    DEFAULT_TIME_FORMAT,
    POSSIBLE_TIME_FORMATS,
)
from ..gpx_elements import Gpx


class Parser:
    """
    File parser.
    """

    def __init__(
        self, source: str | Path | IO[str] | IO[bytes] | bytes, name_spaces: Dict = None
    ) -> None:
        """
        Initialise Parser instance.

        Args:
            source (str | Path | IO[str] | IO[bytes] | bytes): Path to a
                file or a file-like object to parse.
        """
        self.source: str = source

        self.ele_data: bool = False
        self.time_data: bool = False
        self.precisions: dict = DEFAULT_PRECISION_DICT
        self.time_format: str = DEFAULT_TIME_FORMAT

        self.gpx: Gpx = Gpx(xmlns=name_spaces)

    def _find_precision(self, number: str | None) -> int:
        """
        Find decimal precision of a given number.

        Args:
            number (str | None): Number.

        Returns:
            int: Decimal precision.
        """
        if number is None:
            return DEFAULT_PRECISION
        if "." in number:
            _, decimal = number.split(sep=".")
            return len(decimal)
        return 0

    def _find_time_format(self, time: str | None):
        """
        Find time format used in the source file.

        Args:
            time (str | None): Time.
        """
        self.time_data = time is not None

        for tf in POSSIBLE_TIME_FORMATS:
            try:
                datetime.strptime(time, tf)
                self.time_format = tf
                break
            except ValueError:
                pass
        else:
            warnings.warn(
                """Unknown time format. Default time format will be used uppon
                writting."""
            )



"""
This module contains the TrackSegment class.
"""

from .extensions import Extensions
from .gpx_element import GpxElement
from .waypoint import WayPoint


class TrackSegment(GpxElement):
    """
    trksegType element in GPX file.
    """

    fields = ["trkpt", "extensions"]
    mandatory_fields = []

    def __init__(
        self,
        tag: str = "trkseg",
        trkpt: list[WayPoint] = None,
        extensions: Extensions = None,
    ) -> None:
        """
        Initialise TrackSegment instance.

        Args:
            tag (str, optional): XML tag. Defaults to "trkseg".
            trkpt (list[WayPoint], optional): List of track points. Defaults to None.
            extensions (Extensions, optional): Extensions. Defaults to None.
        """
        self.tag: str = tag
        self.trkpt: list[WayPoint] = [] if trkpt is None else trkpt
        self.extensions: Extensions = extensions



"""
This module contains the GPXParser class.
"""

import io
import warnings
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import IO

from ..gpx_elements import (
    Bounds,
    Copyright,
    Email,
    Extensions,
    Gpx,
    Link,
    Metadata,
    Person,
    Point,
    PointSegment,
    Route,
    Track,
    TrackSegment,
    WayPoint,
)
from .xml_parser import XMLParser


class GPXParser(XMLParser):
    """
    GPX file parser.
    """

    def __init__(
        self,
        source: str | Path | IO[str] | IO[bytes] | bytes,
        xml_schemas: bool = True,
        xml_extensions_schemas: bool = False,
    ) -> None:
        """
        Initialise GPXParser instance.

        Args:
            source (str | Path | IO[str] | IO[bytes] | bytes): Path to a
                file or a file-like object to parse.
            xml_schemas (bool, optional): Toggle schema
                verification during parsing. Defaults to True.
            xml_extensions_schemas (bool, optional): Toggle extensions
                schema verificaton durign parsing. Requires internet connection
                connection and is not guaranted to work. Defaults to False.
        """
        # Bytes object
        if isinstance(source, bytes):
            source = io.BytesIO(source)

        # Initialise XMLParser and parse GPX file
        super().__init__(source, xml_schemas, xml_extensions_schemas)
        self.parse()

    def _find_precisions(self):
        """
        Find decimal precision of any type of value in a GPX file (latitude, elevation...).
        Also find if the GPX file contains elevation data.
        """
        # Point
        track = self.xml_root.findall("trk", self.name_spaces)[0]
        segment = track.findall("trkseg", self.name_spaces)[0]
        point = segment.findall("trkpt", self.name_spaces)[0]

        ele_text = point.findtext("ele", namespaces=self.name_spaces)
        if ele_text is not None:
            self.ele_data = True
        else:
            self.ele_data = False

        self.precisions["lat_lon"] = self._find_precision(point.get("lat"))
        self.precisions["elevation"] = self._find_precision(ele_text)

    def _find_time_element(self) -> str | None:
        """
        Find a time element in GPX file.

        Returns:
            str | None: Time element.
        """
        # Use time from metadata
        metadata = self.xml_root.find("metadata", self.name_spaces)
        if metadata is not None:
            time_str = metadata.findtext("time", namespaces=self.name_spaces)
            if time_str is not None:
                return time_str

        # Use time from track point
        track = self.xml_root.findall("trk", self.name_spaces)[
            0
        ]  # Optimise, load only once??
        segment = track.findall("trkseg", self.name_spaces)[0]
        point = segment.findall("trkpt", self.name_spaces)[0]
        time_str = point.findtext("time", namespaces=self.name_spaces)
        if time_str is not None:
            return time_str

        # No time element at all...
        return None

    def _parse_bounds(self, bounds: ET.Element, tag: str = "bounds") -> Bounds | None:
        """
        Parse boundsType element from GPX file.

        Args:
            bounds (ET.Element): Parsed bounds element.
            tag (str, Optional): XML tag. Defaults to "bounds".

        Returns:
            Bounds | None: Bounds instance.
        """
        if bounds is None:
            return None

        return Bounds(
            tag,
            bounds.get("minlat"),
            bounds.get("minlon"),
            bounds.get("maxlat"),
            bounds.get("maxlon"),
        )

    def _parse_copyright(
        self, copyright_: ET.Element, tag: str = "copyright"
    ) -> Copyright | None:
        """
        Parse copyrightType element from GPX file.

        Args:
            copyright (ET.Element): Parsed copyright element.
            tag (str, Optional): XML tag. Defaults to "copyright".

        Returns:
            Copyright | None: Copyright instance.
        """
        if copyright_ is None:
            return None

        return Copyright(
            tag,
            copyright_.get("author"),
            copyright_.findtext("year", namespaces=self.name_spaces),
            copyright_.findtext("licence", namespaces=self.name_spaces),
        )

    def _parse_email(self, email: ET.Element, tag: str = "email") -> Email | None:
        """
        Parse emailType element from GPX file.

        Args:
            email (ET.Element): Parsed email element.
            tag (str, Optional): XML tag. Defaults to "email".

        Returns:
            Email | None: Email instance.
        """
        if email is None:
            return None

        return Email(tag, email.get("id"), email.get("domain"))

    def _parse_extensions(
        self, extensions: ET.Element, element_type: str, tag: str = "extensions"
    ) -> Extensions | None:
        """
        Parse extensionsType element from GPX file.

        Args:
            extensions (ET.Element): Parsed extensions element.
            tag (str, Optional): XML tag. Defaults to "extensions".

        Returns:
            Extensions | None: Extensions instance.
        """
        if extensions is None:
            return None

        def construct_dict(e0):
            e1s = list(e0.iter())[1:]
            if len(e1s) > 0:
                d = {"attrib": dict(e0.items()), "elmts": {}}
                for e1 in e1s:
                    d["elmts"][e1.tag] = construct_dict(e1)
                return d
            return {"attrib": {}, "elmts": e0.text}

        ext = list(extensions.iter())[1]
        values = {ext.tag: {}}
        values[ext.tag] = construct_dict(ext)

        # Etensions fields are based on the first occurance of a type encountered in the file
        if self.extensions_fields.get(element_type) is None:
            self.extensions_fields[element_type] = values

        return Extensions(tag, values)

    def _parse_link(self, link: ET.Element, tag: str = "link") -> Link | None:
        """
        Parse linkType element from GPX file.

        Args:
            link (ET.Element): Parsed link element.
            tag (str, Optional): XML tag. Defaults to "link".

        Returns:
            Link | None: Link instance.
        """
        if link is None:
            return None

        return Link(
            tag,
            link.get("href"),
            link.findtext("text", namespaces=self.name_spaces),
            link.findtext("type", namespaces=self.name_spaces),
        )

    def _parse_metadata(
        self, metadata: ET.Element, tag: str = "metadata"
    ) -> Metadata | None:
        """
        Parse metadataType element from GPX file.

        Args:
            metadata (ET.Element): Parsed metadata element.
            tag (str, Optional): XML tag. Defaults to "metadata".

        Returns:
            Metadata | None: Metadata instance.
        """
        if metadata is None:
            return None

        return Metadata(
            tag,
            metadata.findtext("name", namespaces=self.name_spaces),
            metadata.findtext("desc", namespaces=self.name_spaces),
            self._parse_person(metadata.find("author", self.name_spaces)),
            self._parse_copyright(metadata.find("copyright", self.name_spaces)),
            self._parse_link(metadata.find("link", self.name_spaces)),
            self.find_time(metadata, "time"),
            metadata.findtext("keywords", namespaces=self.name_spaces),
            self._parse_bounds(metadata.find("bounds", self.name_spaces)),
            self._parse_extensions(metadata.find("extensions", self.name_spaces), tag),
        )

    def _parse_person(self, person: ET.Element, tag: str = "person") -> Person | None:
        """
        Parse personType element from GPX file.

        Args:
            person (ET.Element): Parsed person element.
            tag (str, Optional): XML tag. Defaults to "person".

        Returns:
            Person | None: Person instance.
        """
        if person is None:
            return None

        return Person(
            tag,
            person.findtext("name", namespaces=self.name_spaces),
            self._parse_email(person.find("email", self.name_spaces)),
            self._parse_link(person.find("link", self.name_spaces)),
        )

    def _parse_point_segment(
        self, point_segment: ET.Element, tag: str = "ptseg"
    ) -> PointSegment | None:
        """
        Parse ptsegType element from GPX file.

        Args:
            point_segment (ET.Element): Parsed point segment element.
            tag (str, optional): XML tag. Defaults to "ptseg".

        Returns:
            PointSegment | None: PointSegment instance.
        """
        if point_segment is None:
            return None

        return PointSegment(
            tag, [self._parse_point(p) for p in point_segment.findall("pt")]
        )

    def _parse_point(self, point: ET.Element, tag: str = "pt") -> Point | None:
        """
        Parse ptType element from GPX file.

        Args:
            point (ET.Element): Parsed point element.
            tag (str, Optional): XML tag. Defaults to "pt".

        Returns:
            Point | None: Point instance.
        """
        if point is None:
            return None

        return Point(
            tag,
            self.get_float(point, "lat"),
            self.get_float(point, "lon"),
            self.find_float(point, "ele"),
            self.find_time(point, "time"),
        )

    def _parse_route(self, route: ET.Element, tag: str = "rte") -> Route | None:
        """
        Parse rteType element from GPX file.

        Args:
            route (ET.Element): Parsed route element.
            tag (str, Optional): XML tag. Defaults to "rte".

        Returns:
            Route | None: Route instance.
        """
        if route is None:
            return None

        return Route(
            tag,
            route.findtext("name", namespaces=self.name_spaces),
            route.findtext("cmt", namespaces=self.name_spaces),
            route.findtext("desc", namespaces=self.name_spaces),
            route.findtext("src", namespaces=self.name_spaces),
            self._parse_link(route.find("link", self.name_spaces)),
            self.find_int(route, "number"),
            route.findtext("type", namespaces=self.name_spaces),
            self._parse_extensions(route.find("extensions", self.name_spaces), tag),
            [
                self._parse_waypoint(waypoint)
                for waypoint in route.findall("rtept", self.name_spaces)
            ],
        )

    def _parse_track_segment(
        self, track_segment: ET.Element, tag: str = "trkseg"
    ) -> TrackSegment | None:
        """
        Parse trksegType element from GPX file.

        Args:
            track_segment (ET.Element): Parsed track
                segment element.
            tag (str, Optional): XML tag. Defaults to "trkseg".

        Returns:
            TrackSegment | None: TrackSegment instance.
        """
        if track_segment is None:
            return None

        return TrackSegment(
            tag,
            [
                self._parse_waypoint(track_point, "trkpt")
                for track_point in track_segment.findall("trkpt", self.name_spaces)
            ],
            self._parse_extensions(
                track_segment.find("extensions", self.name_spaces), tag
            ),
        )

    def _parse_track(self, track: ET.Element, tag: str = "trk") -> Track | None:
        """
        Parse trkType element from GPX file.

        Args:
            track (ET.Element): Parsed track element.
            tag (str, Optional): XML tag. Defaults to "trk".

        Returns:
            Track | None: Track instance.
        """
        if track is None:
            return None

        return Track(
            tag,
            track.findtext("name", namespaces=self.name_spaces),
            track.findtext("cmt", namespaces=self.name_spaces),
            track.findtext("desc", namespaces=self.name_spaces),
            track.findtext("src", namespaces=self.name_spaces),
            self._parse_link(track.find("link", self.name_spaces)),
            self.find_int(track, "number"),
            track.findtext("type", namespaces=self.name_spaces),
            self._parse_extensions(track.find("extensions", self.name_spaces), tag),
            [
                self._parse_track_segment(segment)
                for segment in track.findall("trkseg", self.name_spaces)
            ],
        )

    def _parse_waypoint(
        self, waypoint: ET.Element, tag: str = "wpt"
    ) -> WayPoint | None:
        """
        Parse wptType element from GPX file.

        Args:
            waypoint (ET.Element): Parsed waypoint element.
            tag (str, Optional): XML tag. Defaults to "wpt".

        Returns:
            WayPoint | None: WayPoint instance.
        """
        if waypoint is None:
            return None

        return WayPoint(
            tag,
            self.get_float(waypoint, "lat"),
            self.get_float(waypoint, "lon"),
            self.find_float(waypoint, "ele"),
            self.find_time(waypoint, "time"),
            self.find_float(waypoint, "magvar"),
            self.find_float(waypoint, "geoidheight"),
            waypoint.findtext("name", namespaces=self.name_spaces),
            waypoint.findtext("cmt", namespaces=self.name_spaces),
            waypoint.findtext("desc", namespaces=self.name_spaces),
            waypoint.findtext("src", namespaces=self.name_spaces),
            self._parse_link(waypoint.find("link", self.name_spaces)),
            waypoint.findtext("sym", namespaces=self.name_spaces),
            waypoint.findtext("type", namespaces=self.name_spaces),
            waypoint.findtext("fix", namespaces=self.name_spaces),
            self.find_int(waypoint, "sat"),
            self.find_float(waypoint, "hdop"),
            self.find_float(waypoint, "vdop"),
            self.find_float(waypoint, "pdop"),
            self.find_float(waypoint, "ageofgpsdata"),
            self.find_float(waypoint, "dgpsid"),
            self._parse_extensions(waypoint.find("extensions", self.name_spaces), tag),
        )

    def _parse_root_properties(self):
        """
        Parse XML properties from GPX file.
        """
        self.gpx.creator = self.xml_root.attrib["creator"]
        self.gpx.version = self.xml_root.attrib["version"]
        schema_location = self.xml_root.get(
            "{http://www.w3.org/2001/XMLSchema-instance}schemaLocation"
        ).split(" ")
        self.gpx.xsi_schema_location = [x for x in schema_location if x != ""]

    def _parse_root_metadata(self):
        """
        Parse metadataType elements from GPX file.
        """
        self.gpx.metadata = self._parse_metadata(
            self.xml_root.find("metadata", self.name_spaces)
        )

    def _parse_root_waypoints(self):
        """
        Parse wptType elements from GPX file.
        """
        waypoints = self.xml_root.findall("wpt", self.name_spaces)
        for waypoint in waypoints:
            self.gpx.wpt.append(self._parse_waypoint(waypoint))

    def _parse_root_routes(self):
        """
        Parse rteType elements from GPX file
        """
        routes = self.xml_root.findall("rte", self.name_spaces)
        for route in routes:
            self.gpx.rte.append(self._parse_route(route))

    def _parse_root_tracks(self):
        """
        Parse trkType elements from GPX file.
        """
        tracks = self.xml_root.findall("trk", self.name_spaces)
        for track in tracks:
            self.gpx.trk.append(self._parse_track(track))

    def _parse_root_extensions(self):
        """
        Parse extensionsType elements from GPX file.
        """
        extensions = self.xml_root.find("extensions", self.name_spaces)
        self.gpx.extensions = self._parse_extensions(extensions, "gpx")

    def parse(self) -> Gpx:
        """
        Parse GPX file.

        Returns:
            Gpx: Gpx instance.
        """
        # Parse GPX file
        try:
            self.xml_tree = ET.parse(self.source)
            self.xml_root = self.xml_tree.getroot()
        except Exception as err:
            warnings.warn(f"Unexpected {err}, {type(err)}.\nUnable to parse GPX file.")
            raise

        # Parse properties
        try:
            self._parse_root_properties()
        except Exception as e:
            raise ValueError("Unable to parse properties in GPX file.") from e

        # Check XML schemas
        self.xml_schemas()

        # Find precisions
        self._find_precisions()

        # Find time format
        self._find_time_format(self._find_time_element())

        # Parse metadata
        try:
            self._parse_root_metadata()
        except:
            warnings.warn("Unable to parse metadata in GPX file.")
            raise

        # Parse way points
        try:
            self._parse_root_waypoints()
        except:
            warnings.warn("Unable to parse waypoints in GPX file.")
            raise

        # Parse routes
        try:
            self._parse_root_routes()
        except:
            warnings.warn("Unable to parse routes in GPX file.")
            raise

        # Parse tracks
        try:
            self._parse_root_tracks()
        except:
            warnings.warn("Unable to parse tracks in GPX file.")
            raise

        # Parse extensions
        try:
            self._parse_root_extensions()
        except:
            warnings.warn("Unable to parse extensions in GPX file.")
            raise

        return self.gpx



"""
This module contains utility functions for distances.
"""

import math as m
import warnings
from typing import Protocol, Tuple

# latitude/longitude in GPX files is always in WGS84 datum
# WGS84 defined the Earth semi-major axis with 6378.137 km
# https://en.wikipedia.org/wiki/World_Geodetic_System#WGS84
EARTH_RADIUS = 6_378_137


class HasLatLon(Protocol):
    """Protocol for objects that have latitude and longitude."""

    lat: float
    lon: float


def haversine_distance(point_1: HasLatLon, point_2: HasLatLon) -> float:
    """
    Compute Haversine distance (in meters) between to points.
    Source: https://en.wikipedia.org/wiki/Haversine_formula

    Args:
        point_1 (HasLatLon): First point.
        point_2 (HasLatLon): Second point.

    Returns:
        float: Haversine distance between the points.
    """
    # Delta and conversion to radians
    delta_lat = m.radians(point_1.lat - point_2.lat)
    delta_long = m.radians(point_1.lon - point_2.lon)

    sin_1 = m.sin(delta_lat / 2)
    sin_2 = m.sin(delta_long / 2)
    a = m.sqrt(
        sin_1 * sin_1
        + m.cos(m.radians(point_1.lat)) * m.cos(m.radians(point_2.lat)) * sin_2 * sin_2
    )
    d = 2 * EARTH_RADIUS * m.asin(a)

    return d


def perpendicular_distance(
    start_point: HasLatLon, end_point: HasLatLon, point: HasLatLon
) -> float:
    """
    Compute perpendicular distance between a point and a line.

    Args:
        start_point (HasLatLon): A point on the line.
        end_point (HasLatLon): A point on the line.
        point (HasLatLon): A point to measure the distance from.

    Returns:
        float: Perpendicular distance between the point *point* and the
            line defined by *start_point* and *end_point*.
    """

    def line_coefficients(
        point_1: HasLatLon, point_2: HasLatLon
    ) -> Tuple[float, float, float]:
        """
        Compute the coefficients of a line equation of the form: ax+by+c=0.

        Args:
            point_1 (HasLatLon): A point on the line.
            point_2 (HasLatLon): A point on the line.
        Returns:
            Tuple[float, float, float]: Coefficients of the line equation.
        """
        delta_x = point_1.lon - point_2.lon
        delta_y = point_1.lat - point_2.lat

        try:
            a = delta_y / delta_x
            b = -1
            c = point_1.lat - a * point_1.lon
        except ZeroDivisionError:
            a = 1
            b = 0
            c = point_1.lon
            warnings.warn("Vertical line")

        return a, b, c

    a, b, c = line_coefficients(start_point, end_point)

    d = abs(a * point.lon + b * point.lat + c) / m.sqrt(a * a + b * b)

    return d



"""
This module contains the Point class.
"""

import datetime

from .gpx_element import GpxElement


class Point(GpxElement):
    """
    ptType element in GPX file.
    """

    fields = ["lat", "lon", "ele", "time"]
    mandatory_fields = ["lat", "lon"]

    def __init__(
        self,
        tag: str = "pt",
        lat: float = None,
        lon: float = None,
        ele: float = None,
        time: datetime = None,
    ) -> None:
        """
        Initialise Point instance.

        Args:
            tag (str, optional): XML tag. Defaults to "pt".
            lat (float, optional): Latitude (in degrees). Defaults to None.
            lon (float, optional): Longitude (in degrees). Defaults to None.
            ele (float, optional): Elevation (in meters). Defaults to None.
            time (datetime, optional): Time. Defaults to None.
        """
        self.tag: str = tag
        self.lat: float = lat
        self.lon: float = lon
        self.ele: float = ele
        self.time: datetime = time



"""
This module contains the Extensions class.
"""

from .gpx_element import GpxElement


class Extensions(GpxElement):
    """
    extensionsType element in GPX file.
    """

    fields = []
    mandatory_fields = []

    def __init__(self, tag: str = "extensions", values: dict = None) -> None:
        """
        Initialise Extensions instance.

        Args:
            tag (str, optional): XML tag. Defaults to "extensions".
            values (dict, optional): Extension content. Defaults to None.
        """
        self.tag: str = tag
        self.values: dict = values if values is not None else {}



"""
This module contains the KMLParser class.
"""

import io
import warnings
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import IO

from ..gpx_elements import Gpx, Track, TrackSegment, WayPoint
from .xml_parser import XMLParser


class KMLParser(XMLParser):
    """
    KML file parser.
    """

    def __init__(
        self,
        source: str | Path | IO[str] | IO[bytes] | bytes,
        xml_schemas: bool = True,
        xml_extensions_schemas: bool = False,
    ) -> None:
        """
        KML file parser.

        Args:
            source (str | Path | IO[str] | IO[bytes] | bytes): Path to a
                file or a file-like object to parse.
            xml_schemas (bool, optional): Toggle schema
                verification during parsing. Defaults to True.
            xml_extensions_schemas (bool, optional): Toggle extensions
                schema verificaton durign parsing. Requires internet connection
                connection and is not guaranted to work. Defaults to False.

        Raises:
            ValueError: Source is not valid.
        """
        # Bytes object
        if isinstance(source, bytes):
            source = io.BytesIO(source)

        # Initialise XMLParser and parse KML file
        super().__init__(source, xml_schemas, xml_extensions_schemas)
        self.parse()

    def find_precisions(self):
        """
        Find decimal precision of any type of value in a KML file (latitude, elevation...).
        """
        # Point
        documents = self.xml_root.findall("Document", self.name_spaces)
        placemarks = documents[0].findall("Placemark", self.name_spaces)
        linestrings = placemarks[0].findall("LineString", self.name_spaces)
        coordinates = self.find_text(linestrings[0], "coordinates")

        coordinates = coordinates.replace("\n", "").replace("\t", "")
        if coordinates[-1] == " ":
            coordinates = coordinates[:-1]
        coordinates = coordinates.split(" ")
        coordinates = coordinates[0].split(",")

        self.precisions["lat_lon"] = self._find_precision(coordinates[0])
        self.precisions["elevation"] = self._find_precision(coordinates[2])

    # def parse_linestring(self, linestring) -> list[str]:
    #     """
    #     Parse LineString element from KML file.

    #     Args:
    #         placemark (ET.Element): Parsed LineString element.

    #     Returns:
    #         list[str]: Informations contained in the LineString element (strings of coordinates).
    #     """
    #     if linestring is None:
    #         return None

    #     linestrings_data = []
    #     coordinatess = linestring.findall("opengis:coordinates", self.name_space)
    #     for coordinates in coordinatess:
    #         linestrings_data.append(coordinates.text)

    #     return linestrings_data

    def parse_placemark(self, placemark) -> dict | None:
        """
        Parse Placemark element from KML file.

        Args:
            placemark (ET.Element): Parsed Placemark
                element.

        Returns:
            dict | None: Informations contained in the Placemark
                element (the name of the Placemark (str) and the
                contents of the LineString (list[str])).
        """
        if placemark is None:
            return None

        placemark_data = {}

        placemark_data["name"] = self.find_text(placemark, "name")

        placemark_data["linestrings_data"] = []
        linestrings = placemark.findall("LineString", self.name_spaces)
        for linestring in linestrings:
            placemark_data["linestrings_data"].append(
                self.find_text(linestring, "coordinates")
            )

        return placemark_data

    def parse_document(self, document) -> list[dict] | None:
        """
        Parse Document element from KML file.

        Args:
            document (ET.Element): Parsed Document element.

        Returns:
            list[dict] | None: Informations related to Placemark
                elements contained in the Document element.
        """
        if document is None:
            return None

        # name = self.find_text(document, "name")

        placemmarks_data = []
        placemarks = document.findall("Placemark", self.name_spaces)
        for placemark in placemarks:
            placemmarks_data.append(self.parse_placemark(placemark))

        return placemmarks_data

    def parse_root_document(self):
        """
        Parse Document elements from KML file.
        """
        documents = self.xml_root.findall("Document", self.name_spaces)
        for document in documents:
            placemarks_data = self.parse_document(document)

            if len(placemarks_data) == 1:
                placemark_data = placemarks_data[0]
                linestrings_data = placemark_data["linestrings_data"]
                trkseg = []
                for coordinates in linestrings_data:
                    coordinates = coordinates.replace("\n", "").replace("\t", "")
                    if coordinates[-1] == " ":
                        coordinates = coordinates[:-1]
                    coordinates = coordinates.split(" ")
                    trkpt = []
                    for point_coord in coordinates:
                        point_coord = point_coord.split(",")
                        trkpt.append(
                            WayPoint(
                                tag="trkpt",
                                lat=float(point_coord[1]),
                                lon=float(point_coord[0]),
                                ele=float(point_coord[2]),
                            )
                        )
                    trkseg.append(TrackSegment(trkpt=trkpt))

                tracks = [Track(name=placemark_data["name"], trkseg=trkseg)]
                self.gpx.trk = tracks
            else:
                warnings.warn("Oops, not yet implemented...")

    def add_properties(self):
        """
        Add missing properties (properties that do not exist in KML/KMZ
        files but mandatory in GPX files).
        """
        self.gpx.creator = "ezGPX"
        self.gpx.xmlns = "http://www.topografix.com/GPX/1/1"
        self.gpx.version = "1.1"
        self.gpx.xmlns_xsi = "http://www.w3.org/2001/XMLSchema-instance"
        self.gpx.xsi_schema_location = [
            "http://www.topografix.com/GPX/1/1",
            "http://www.topografix.com/GPX/1/1/gpx.xsd",
        ]

    def parse(self) -> Gpx:
        """
        Parse KML file.

        Returns:
            Gpx: Gpx instance.
        """
        # Parse KML file
        try:
            self.xml_tree = ET.parse(self.source)
            self.xml_root = self.xml_tree.getroot()
        except Exception as err:
            warnings.warn(
                "Unexpected %s, %s.\nUnable to parse KML file.", err, type(err)
            )
            raise

        # Add properties
        self.add_properties()

        # Find precisions
        self.find_precisions()

        # Check XML schemas
        self.xml_schemas()

        # Parse Document
        try:
            self.parse_root_document()
        except:
            warnings.warn("Unable to parse tracks in GPX file.")
            raise

        return self.gpx



"""
This module contains the PointSegment class.
"""

from .gpx_element import GpxElement
from .point import Point


class PointSegment(GpxElement):
    """
    ptsegType element in GPX file.
    """

    fields = ["pt"]
    mandatory_fields = []

    def __init__(self, tag: str = "ptseg", pt: list[Point] = None) -> None:
        """
        Initialise PointSegment instance.

        Args:
            tag (str, optional): XML tag. Defaults to "ptseg".
            pt (list[Point], optional): List of points. Defaults to None.
        """
        self.tag: str = tag
        self.pt: list[Point] = [] if pt is None else pt



"""
This module contains the XMLParser class.
"""

import warnings
import xml.etree.ElementTree as ET
from datetime import datetime
from pathlib import Path
from typing import IO

from dateutil import parser

from ..utils import check_xml_extensions_schemas, check_xml_schema
from .parser import Parser


class XMLParser(Parser):
    """
    XML file parser.
    """

    def __init__(
        self,
        source: str | Path | IO[str] | IO[bytes] | bytes,
        xml_schema: bool = True,
        xml_extensions_schemas: bool = False,
    ) -> None:
        """
        Initialise XML Parser instance.

        Args:
            source (str | Path | IO[str] | IO[bytes] | bytes): Path to
                a file or a file-like object to parse.
            xml_schema (bool, optional): Toggle  schema verification
                during parsing. Defaults to True.
            xml_extensions_schemas (bool, optional): Toggle extensions
                schema verificaton durign parsing. Requires internet
                connection and is not guaranted to work. Defaults to False.
        """
        self.name_spaces: dict = {
            node[0]: node[1] for _, node in ET.iterparse(source, events=["start-ns"])
        }
        self.extensions_fields: dict = {}

        super().__init__(source, self.name_spaces)

        self.xml_schema: bool = xml_schema
        self.xml_extensions_schemas: bool = xml_extensions_schemas

        self.xml_tree: ET.ElementTree = None
        self.xml_root: ET.Element = None

    def get_text(self, element: ET.Element, sub_element: str) -> str | None:
        """
        Get text from sub-element.

        Args:
            element (ET.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            str | None: Text from sub-element.
        """
        text_ = element.get(sub_element)
        if text_ is None:
            warnings.warn(f"{element} has no attribute {sub_element}.")
        return text_

    def get_int(self, element: ET.Element, sub_element: str) -> int | None:
        """
        Get integer value from sub-element.

        Args:
            element (ET.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            int | None: Integer value from sub-element.
        """
        int_ = element.get(sub_element)
        if int_ is None:
            warnings.warn(f"{element} has no attribute {sub_element}.")
        else:
            int_ = int(int_)
        return int_

    def get_float(self, element: ET.Element, sub_element: str) -> float | None:
        """
        Get floating point value from sub-element.

        Args:
            element (ET.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            float | None: Floating point value from sub-element.
        """
        float_ = element.get(sub_element)
        if float_ is None:
            warnings.warn(f"{element} has no attribute {sub_element}.")
        else:
            float_ = float(float_)
        return float_

    def find_sub_element(
        self, element: ET.Element, sub_element: str
    ) -> ET.Element | None:
        """
        Find sub-element.

        Args:
            element (ET.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            ET.Element | None: Sub-element.
        """
        sub_element_ = element.find(sub_element, self.name_spaces)
        if sub_element_ is None:
            warnings.warn(f"{element} has no attribute {sub_element}.")
        return sub_element_

    def find_text(self, element: ET.Element, sub_element: str) -> str | None:
        """
        Find text from sub-element.

        Args:
            element (ET.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            str | None: Text from sub-element.
        """
        sub_element_ = self.find_sub_element(element, sub_element)
        return None if sub_element_ is None else sub_element_.text

    def find_int(self, element: ET.Element, sub_element: str) -> int | None:
        """
        Find integer value from sub-element.

        Args:
            element (ET.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            int | None: Integer value from sub-element.
        """
        sub_element_ = self.find_sub_element(element, sub_element)
        return None if sub_element_ is None else int(sub_element_.text)

    def find_float(self, element: ET.Element, sub_element: str) -> float | None:
        """
        Find float point value from sub-element.

        Args:
            element (ET.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            float | None: Floating point value from sub-element.
        """
        sub_element_ = self.find_sub_element(element, sub_element)
        return None if sub_element_ is None else float(sub_element_.text)

    def find_time(self, element: ET.Element, sub_element: str) -> datetime | None:
        """
        Find time value from sub-element.

        Args:
            element (ET.Element): Parsed element from GPX file.
            sub_element (str): Sub-element name.

        Returns:
            datetime | None: Floating point value from sub-element.
        """
        sub_element_ = self.find_sub_element(element, sub_element)
        return None if sub_element_ is None else parser.parse(sub_element_.text)

    def xml_schemas(self):
        """
        Check XML schemas during parsing.
        """
        # Check XML schema
        if self.xml_schema:
            if not check_xml_schema(self.source, self.gpx.version):
                raise ValueError("Invalid GPX file (does not follow XML schema).")

        # Check XML extension schemas
        if self.xml_extensions_schemas:
            if not check_xml_extensions_schemas(self.source):
                raise ValueError(
                    "Invalid GPX file (does not follow XML extensions schemas)."
                )
